##dream.S
##
##This is a scheme interpreter written in x86 assembly
##thanks to the description given by Abelson and Sussman in
##Structure and Interpretation of Computer Programs.
##
##Copyright (C) 2004 David Joseph Stith
##
##This program is free software; you can redistribute it and/or modify
##it under the terms of the GNU General Public Licence as published by
##the Free Software Foundation; either version 2 of the License, or
##(at your option) any later version.
##
##This program is distributed in the hope that it will be useful,
##but WITHOUT ANY WARRANTY; without even the implied warranty of
##MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##GNU General Public License for more details.
##
##You should have received a copy of the GNU General Public License
##along with this program; if not, write to the Free Software
##Foundation, Inc.,
##59 Temple Place - Suite 330, Boston MA  02111-1307, USA
##
##To contact the author, email to joey@stripedgazelle.org
##or snail mail to:
##David Joseph Stith
##3133 Meadow Wood Ln.
##Bedford, TX 76021
##------------------------------------------------------------------------------
.globl _start
.globl dream_error
.globl dream_free
.globl dream_memlimit
.globl dream_true
.globl dream_false
.globl error_too_many_args
.globl error_too_few_args
.globl error_expected_integer
.globl error_expected_string

##-----------------------------------
##Continuation control stack register
##-----------------------------------
#define SP      %esp

##---------------------
##Interpreter registers
##---------------------
#define    EXP     %edx
#define    ENV     %ebp
#define    UNEV    %esi
#define    ARGL    %edi
#define    VAL     %eax
#define    VALH    %ah
#define    VALL    %al
#define    FREE    %ebx
#define    TEMP    %ecx
#define    TEMPW   %cx
#define    TEMPH   %ch
##
##---------------------------
##Garbage collector registers
##---------------------------
#define    OLD     %esi
#define    NEW     %edi
#define    SCAN    %eax

##------------------------------------------------------
##Object types all must be incongruent to 0 mod 2
##to distinguish them from the car field of pairs.
##------------------------------------------------------
##Negative types;are not copied by the garbage collector
##------------------------------------------------------
.equ    SYMBOL, -9
.equ    MEMOIZABLE_SYMBOL, SYMBOL - (1 * 256)

##----------
##Characters
##----------
.equ    CHAR, -7 #High byte holds character value


##Boolean types
.equ    BOOLEAN, -5
.equ    BOOLEAN_TRUE, BOOLEAN - (1 * 256)
.equ    BOOLEAN_FALSE, BOOLEAN - (0 * 256)

##-----------
##End of File
##-----------
.equ    EOF, -3

##---------------------------------------------------------------------
## BROKEN_HEART is used during garbage collection as a forwarding addr.
##---------------------------------------------------------------------
.equ    BROKEN_HEART, -1

##---------------------------------------------------------------------------
##Positive types are copied by the garbage collector, and 
##if (test type, 256) is non-zero then cdr field is copied as well.
##---------------------------------------------------------------------------

##----------
##Procedures
##----------
.equ    PROCEDURE, 1

## A SPECIAL_FORM is a pointer
.equ    SPECIAL_FORM, PROCEDURE + (0 * 256)

## A PRIMITIVE is a pointer
.equ    PRIMITIVE, PROCEDURE + (2 * 256)
##
## A CLOSURE is (environment formals . body)
## where environment is (bindings . enclosing_scope)
## where bindings are ((symbol . value) ...)
.equ    CLOSURE, PROCEDURE + (1 * 256)

## A CONTINUATION is (root . stack)
## where stack is (special_form ...)
.equ    CONTINUATION, PROCEDURE + (3 * 256)

## A FOREIGN_PROCEDURE is a pointer
.equ    FOREIGN_PROCEDURE, PROCEDURE + (4 * 256)


##-------
##Numbers
##-------
.equ    NUMBER, 3
##The following values must increase as the numeric complexity increases.
.equ    INTEGER, 3 + (0 * 256)       #High word is 1 if INTEGER is inexact.
##A RATIONAL is (numerator . denominator)
##where numerator and denominator are both integers
##so that the value is produced by (/ numerator denominator)
.equ    RATIONAL, 3 + (1 * 256)
.equ    REAL, 3 + (3 * 256)
##A COMPLEX is (real-part . imaginary-part)
.equ    COMPLEX, 3 + (5 * 256)

##-----
##Ports
##-----
.equ    INPUT_PORT, 5 #Low byte of high word holds peeked char
.equ    OUTPUT_PORT, 7

##-------
##Strings
##-------
.equ    STRING, 9         #High word holds length (up to 32767)
.equ    MUTABLE_STRING, 9 + (0 * 256)
.equ    IMMUTABLE_STRING, 9 + (2 * 256)

##-------
##Vectors
##-------
.equ    VECTOR, 11 + (1 * 256)  #High word holds length (up to 32767)

##------------------------------------------------------------------------------
.text
.macro stack_push reg
        movl \reg, (FREE)       #(set-car! FREE reg)
        movl root, TEMP
        movl TEMP, 4(FREE)   #(set-cdr! FREE TEMP)
        movl FREE, root         #(set root (cons reg root))
        addl $8, FREE
.endm

.macro save reg
        stack_push \reg
        cmpl memlimit, FREE
.ifndef DEBUG
        jl no_gc\@
.endif
        call gc_sans_strings
no_gc\@:
.endm

.macro restore reg
        movl root, TEMP
        movl (TEMP), \reg       #reg <- (car TEMP)
        movl 4(TEMP), TEMP      #TEMP <- (cdr TEMP)
        movl TEMP, root         #(set root (cdr root))
.endm

.macro to_lower byte_reg
    #Standard case is lower-case.
        cmpb $'A', \byte_reg
        jb standard_case\@
        cmpb $'Z', \byte_reg
        ja standard_case\@
        addb $32, \byte_reg
    standard_case\@:
.endm
##------------------------------------------------------------------------------
##START
##------------------------------------------------------------------------------
_start: 
        movl $0, (dream_error)
##-------------------------
##INITIALIZE GLOBAL SYMBOLS
##-------------------------
.ifdef RANDOM
        pushl $0
        call time
        movl %eax, (SP)
        call srand
        addl $4, SP
.endif
        movl mem, FREE
        movl $obhash, ARGL
        movl $sym_else, VAL             #cond needs this symbol
        call intern_new_symbol
        movl $sym_arrow, VAL            #cond needs this symbol
        call intern_new_symbol
        movl $sym_dot, VAL              #the list reader needs this symbol
        call intern_new_symbol
        movl $sym_eol, VAL              #the list reader needs this symbol
        call intern_new_symbol
        movl $sym_space_name, VAL       #the char reader needs this symbol
        call intern_new_symbol
        movl $sym_unquote, VAL          #quasiquote needs this symbol
        call intern_new_symbol
        movl $sym_unquote_splicing, VAL #quasiquote needs this symbol
        call intern_new_symbol

##-----------------------------------------
##INITIALIZE DEFAULT INPUT AND OUTPUT PORTS
##-----------------------------------------
        movl FREE, current_input_port
        movl $INPUT_PORT, (FREE)
        movl stdin, TEMP
        movl TEMP, 4(FREE)
        addl $8, FREE
        movl FREE, current_output_port
        movl $OUTPUT_PORT, (FREE)
        movl stdout, TEMP
        movl TEMP, 4(FREE)
        addl $8, FREE
        movb $10, input

##-----------------------------
##INITIALIZE GLOBAL ENVIRONMENT
##-----------------------------
        xorl ENV, ENV   #In this we will form global bindings.
        movl $globals, UNEV
        jmp install_global_loop_begin
    install_global_loop:
        call intern_new_symbol
        movl VAL, 8(FREE)
        lodsl
        movl VAL, (FREE) #Type
        lodsl
        movl VAL, 4(FREE) #Data
        movl FREE, 12(FREE)     #value
        addl $8, FREE
        movl FREE, 8(FREE)      #((symbol . value) . ENV)
        movl ENV, 12(FREE)
        addl $8, FREE
        movl FREE, ENV
        addl $8, FREE
    install_global_loop_begin:
        lodsl
        testl VAL, VAL
        jnz install_global_loop

        movl $memstr1, freestring
        movl ENV, (FREE)
        movl FREE, ENV
        movl $0, 4(ENV)           #ENV <- (cons UNEV ())
        addl $8, FREE
        movl ENV, global_environment

##--------------------
##INITIALIZE CHARARRAY
##--------------------
        movl $chararray, ARGL
        movl $CHAR, VAL
        xorb VALH, VALH
chararray_loop: 
        stosl
        incb VALH
        jne chararray_loop

##---------------------------
##INITIALIZE SCHEME REGISTERS
##---------------------------
        xorl VAL, VAL
        xorl EXP, EXP
        xorl ARGL, ARGL
        xorl UNEV, UNEV

##--------------------------
##LOAD INITIALIZATION SCHEME
##--------------------------
        movl SP, stackbase
        movl $parse_command_line, error_continuation
        movl $init_scheme, string_port_pointer
        movl $string_port, %edi
        call load_scheme_argl
    parse_command_line: 
        movl stackbase, TEMP
        addl $8, TEMP
        cmpl $0, (TEMP)
        jz welcome
        movl $exit_not_ok, error_continuation
        movl (TEMP), TEMP
        call prim_load_scratch
##------------------------------------------------------------------------------
##WELCOME
##------------------------------------------------------------------------------
welcome: 
        call set_output_port_current
        movl $msg_welcome, TEMP
        call puts
        movl $read_eval_print_loop, error_continuation
##------------------------------------------------------------------------------
##INTERACTIVE LOOP
##------------------------------------------------------------------------------
read_eval_print_loop: 
        movl $prompt, TEMP
        call puts
        call set_input_port_current
        call prim_read_argl
        cmpl $eof, VAL
        je exit_ok
        cmpl $sym_eol, VAL
        je error_extra_eol
        movl VAL, EXP
        movl global_environment, ENV
        call eval_dispatch
        call set_output_port_current
        call write_dispatch
        jmp read_eval_print_loop
##------------------------------------------------------------------------------
exit_ok: 
        xorl FREE, FREE
        jmp sys_exit
##------------------------------------------------------------------------------
##OS INTERFACE
##------------------------------------------------------------------------------
sys_exit: #Exit code is in ebx
        pushl %ebx
        call exit
##------------------------------------------------------------------------------
getch:   #Reads one character from file given in io_file (long) and places the
         #character in input (byte).
        pusha
        cmpl $string_port_pointer, io_file
        je getch_string_port
        pushl io_file
        call fgetc
        addl $4, %esp
        cmpl $-1, %eax
        jne getch_end
        movb $0, %al
    getch_end:
        movb %al, input
        popa
        ret

    getch_string_port:
        movl string_port_pointer, %eax
        movb (%eax), %al
        testb %al, %al
        jz getch_end
        incl string_port_pointer
        jmp getch_end
##------------------------------------------------------------------------------
putch:   #Displays character in output (byte) to file given in io_file (long).
        pusha
        pushl io_file
        movzbl output, %ecx
        pushl %ecx
        call fputc
        addl $8, %esp
        test %eax, %eax
        js error_io
        popa
        ret
##------------------------------------------------------------------------------
puts:    #Displays string in ecx to file given in io_file (long).
        pusha
        pushl %ecx
        pushl $percent_s
        pushl io_file
        call fprintf
        addl $12, %esp
        test %eax, %eax
        js error_io
        popa
        ret
##------------------------------------------------------------------------------
puts_escape:  #Writes string in ecx to file given in io_file (long).
        pusha
        pushl io_file
        subl $4, %esp
        movl %ecx, %esi
        jmp puts_escape_loop_begin

    puts_escape_loop: 
        cmpb $0x22, %bl       #double quote
        je puts_escape_backslash
        cmpb $'\\', %bl
        je puts_escape_backslash
    puts_escape_continue: 
        movl %ebx, (%esp)
        call fputc
        incl %esi
    puts_escape_loop_begin: 
        movzbl (%esi), %ebx
        testb %bl, %bl
        jnz puts_escape_loop
        addl $8, %esp
        popa
        ret

    puts_escape_backslash: 
        movb $'\\', (%esp) 
        call fputc
        jmp puts_escape_continue
##------------------------------------------------------------------------------
open_file: #Opens file whose path is ecx using flags in io_file (long).
           #Places descriptor in io_file.
        pusha
        pushl io_file
        pushl %ecx
        call fopen
        addl $8, %esp
        testl %eax, %eax
        jz error_io
        movl %eax, io_file
        popa
        ret
##------------------------------------------------------------------------------
close_file:  #Closes file whose descriptor is ecx
        pusha
        cmp $-1, %ecx
        je close_file_already
        pushl %ecx
        call fclose
        addl $4, %esp
        testl %eax, %eax
        jnz error_io
    close_file_already:
        popa
        ret
##------------------------------------------------------------------------------
prim_system:
        pusha
        testl ARGL, ARGL
        jz prim_system_call
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        call get_last_string
        movl 4(ARGL), ARGL
    prim_system_call:
        push ARGL
        call system
        addl $4, %esp
        movl %eax, io_file
        popa
        movl $INTEGER, (FREE)
        movl (io_file), VAL
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free
.ifdef RANDOM
##------------------------------------------------------------------------------
prim_random:
        call get_last_exact_integer
        movl 4(TEMP), TEMP
        pusha
        call rand
        movl %eax, io_file
        popa
        cltd
        movl (io_file), VAL
        idivl TEMP
        testl EXP, EXP
        jge prim_random_positive
        addl TEMP, EXP
    prim_random_positive:
        movl $INTEGER, (FREE)
        movl EXP, 4(FREE)
        movl FREE, VAL
        xorl EXP, EXP
        jmp advance_free
.endif
##------------------------------------------------------------------------------
##INTERNAL UTILITIES
##------------------------------------------------------------------------------
intern_new_symbol: ##Interns new symbol VAL in hashtable at ARGL
                   ##VAL *MUST* be a NEW symbol
        movb 4(VAL), TEMP
        andl $0x7f, TEMP
        shll $HASHPOWER, TEMP
        addl ARGL, TEMP
    intern_new_symbol_loop:
        cmpl $0, 0(TEMP)
        je intern_new_symbol_found
        addl $4, TEMP
        jmp intern_new_symbol_loop
    intern_new_symbol_found:
        movl VAL, 0(TEMP)
        ret
##------------------------------------------------------------------------------
advance_free: 
        addl $8, FREE
        cmpl memlimit, FREE
        jge gc_sans_strings
        ret
##------------------------------------------------------------------------------
insure_string_space:    #Insures that there are TEMP bytes available
        pushl TEMP
        testl TEMP, TEMP
        js error_out_of_memory
        addl freestring, TEMP
        cmpl memstrlimit, TEMP
        jl insure_string_space_ok
        call gc_strings
        movl (SP), TEMP
        addl freestring, TEMP
        cmpl memstrlimit, TEMP
        jg error_out_of_memory
    insure_string_space_ok:
        popl TEMP
        ret
##------------------------------------------------------------------------------
stringify_number:    #Returns in TEMP the string value of VAL using [radix].
                        #Destroys VAL, ARGL, and EXP in the process.
        testl VAL, VAL
        jns stringify_number_abs
        negl VAL
        jo error_overflow
        call stringify_number_abs
        decl TEMP
        movb $'-', (TEMP)
        ret

    stringify_number_abs: 
        movl $outputend, ARGL
        movb $0, (ARGL)
    stringify_number_loop: 
        decl ARGL
        xorl EXP, EXP
        divl radix
        addw $'0', EXP
        cmpw $'9', EXP
        jle stringify_number_digit
        addw $39, EXP    #Advance to letters for digits greater than 9
    stringify_number_digit: 
        movb EXP, (ARGL)
        testl VAL, VAL
        jnz stringify_number_loop

    stringify_number_end: 
        movl ARGL, TEMP
        xorl VAL, VAL    #Clear scheme register
        xorl ARGL, ARGL    #Clear scheme register
        xorl EXP, EXP    #Clear scheme register
        ret
##------------------------------------------------------------------------------
    digitize_stream_negative:
        call *thunk
        cmpb $0, input
        je digitize_stream_error
        call digitize_stream_absolute
        negl VAL
        jo digitize_stream_error
        ret

digitize_stream:   #returns value in VAL obtained from converting, according
                   #to [radix], the ascii digits streaming into the byte at
                   #[input] from calls to [thunk].
                   #0 is returned in EXP if successful.
        cmpb $0, input
        je digitize_stream_error        #Empty stream
digitize_stream_nonempty:
        cmpb $'-', input
        je digitize_stream_negative
digitize_stream_absolute: 
        xorl VAL, VAL
    digitize_stream_loop: 
        movzbl input, TEMP
        cmpb $0, TEMP
        je digitize_stream_end
        cmpb $' ', TEMP
        je digitize_stream_end
        cmpb $10, TEMP          #newline
        je digitize_stream_end
        cmpb $9, TEMP           #tab
        je digitize_stream_end
        cmpb $'/', TEMP
        je digitize_stream_end
        cmpb $'(', TEMP
        je digitize_stream_end
        cmpb $')', TEMP
        je digitize_stream_end
        cmpb $'"', TEMP
        je digitize_stream_end
        cmp $'0', TEMP
        jb digitize_stream_error
        subb $'0', TEMP
        cmpb $9, TEMP
        ja digitize_stream_alpha_digit
    digitize_stream_digit: 
        imull radix
        jo digitize_stream_error
        cmpl radix, TEMP
        jae digitize_stream_error
        addl TEMP, VAL
        jo digitize_stream_error
        call *thunk
        jmp digitize_stream_loop

    digitize_stream_error: 
        movl $1, EXP
        ret

    digitize_stream_alpha_digit: 
        cmpb $49, TEMP    #adjusted 'a'
        jae digitize_stream_lower_alpha_digit
        cmpb $17, TEMP    #adjusted 'A'
        jb digitize_stream_error
        subb $7, TEMP
        jmp digitize_stream_digit

    digitize_stream_lower_alpha_digit: 
        subb $39, TEMP
        jmp digitize_stream_digit

    digitize_stream_end: 
        xorl EXP, EXP  #Clear scheme register
        ret
##------------------------------------------------------------------------------
##GARBAGE COLLECTION
##------------------------------------------------------------------------------
gc_strings: 
        pushl memstrnew
        jmp gc

gc_sans_strings: 
        pushl $0
    gc: 
        stack_push EXP
        stack_push ENV
        stack_push VAL
        stack_push ARGL
        stack_push UNEV
        movl global_environment, VAL
        stack_push VAL
        movl current_input_port, VAL
        stack_push VAL
        movl current_output_port, VAL
        stack_push VAL
        popl TEMP

        movl memnew, FREE
        movl memnew, SCAN
        movl root, OLD
        call relocate_old_result_in_new
        movl NEW, root
    gc_loop: 
        cmpl FREE, SCAN
        je gc_flip
        movl (SCAN), OLD       #OLD <- (car SCAN)
        testl $1, OLD
        jnz gc_non_pair
        testl OLD, OLD
        jz gc_relocate_cdr
        call relocate_old_result_in_new
        movl NEW, (SCAN)       #(set-car! SCAN NEW)
    gc_relocate_cdr: 
        movl 4(SCAN), OLD      #OLD <- (cdr SCAN)
        testl OLD, OLD
        jz gc_next
        call relocate_old_result_in_new
        movl NEW, 4(SCAN)      #(set-cdr! SCAN NEW)
    gc_next: 
        addl $8, SCAN
        jmp gc_loop

    gc_non_pair: 
        cmpl $0, OLD
        jl gc_next
        testl $256, OLD
        jnz gc_relocate_cdr
        jecxz gc_next
        cmpw $MUTABLE_STRING, OLD
        jne gc_next
    gc_relocate_string: 
        movl 4(SCAN), OLD      #OLD now points to original string
        movl TEMP, NEW         #NEW is new
        movl NEW, 4(SCAN)
        cld
    gc_relocate_string_loop: 
        movsb
        cmpb $0, -1(OLD)
        jne gc_relocate_string_loop
        movl NEW, TEMP
        jmp gc_next

    gc_flip: 
        movl mem, OLD
        movl memnew, NEW
        movl NEW, mem
        movl OLD, memnew           #mem <-> memnew
        movl memlimit, OLD
        movl memlimitnew, NEW
        movl NEW, memlimit
        movl OLD, memlimitnew      #memlimit <-> memlimitnew
        cmpl NEW, FREE
        jge error_out_of_memory
        jecxz gc_flip_ignore_strings
        movl memstr, OLD
        movl memstrnew, NEW
        movl NEW, memstr
        movl OLD, memstrnew     #memstr <-> memstrnew
        movl memstrlimit, OLD
        movl memstrlimitnew, NEW
        movl NEW, memstrlimit
        movl OLD, memstrlimitnew        #memstrlimit <-> memstrlimitnew
        movl TEMP, freestring
    gc_flip_ignore_strings: 
        restore VAL
        movl VAL, current_output_port
        restore VAL
        movl VAL, current_input_port
        restore VAL
        movl VAL, global_environment
        restore UNEV
        restore ARGL
        restore VAL
        restore ENV
        restore EXP
        ret
##------------------------------------------------------------------------------
relocate_old_result_in_new: 
        movl (OLD), EXP        #EXP <- (car OLD)
        testl $1, EXP
        je relocate_move
        cmpl $BROKEN_HEART, EXP
        je relocate_already_moved
        cmpb $0, EXP
        jl relocate_do_not_move
        cmpw $VECTOR, EXP
        je relocate_vector
    relocate_move: 
        movl FREE, NEW
        addl $8, FREE
        movl EXP, (NEW)        #(set-car! NEW EXP)
        movl 4(OLD), ENV       #ENV <- (cdr OLD)
        movl ENV, 4(NEW)       #(set-cdr! NEW ENV)
        movl $BROKEN_HEART, (OLD) #(set-car! OLD BROKEN_HEART)
        movl NEW, 4(OLD)       #(set-cdr! OLD NEW)
        ret

    relocate_do_not_move: 
        movl OLD, NEW
        ret

    relocate_already_moved: 
        movl 4(OLD), NEW
        ret

    relocate_vector: 
        call relocate_move
        pushl NEW
        movl FREE, NEW
        addl $8, OLD
        shrl $16, EXP   #Now dx is length of vector
        testw EXP, EXP
        jz relocate_vector_end
        testw $1, EXP    #We always store remaining vector elements in pairs.
        jnz relocate_vector_odd
    relocate_vector_loop: 
        movsl
    relocate_vector_odd: 
        decw EXP
        ja relocate_vector_loop
    relocate_vector_end: 
        movl NEW, FREE
        popl NEW
        ret
##------------------------------------------------------------------------------
##ERRORS AND TYPE VALIDATION
##------------------------------------------------------------------------------
error_io: 
        popa
error_scheme_io:
        movl $msg_io, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_too_many_args: 
        movl $msg_too_many_args, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_too_few_args: 
        movl $msg_too_few_args, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
insure_one_last_arg:
        testl ARGL, ARGL
        jz error_too_few_args
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        ret
##------------------------------------------------------------------------------
error_undefined_var: 
        movl EXP, VAL
        movl stderr, TEMP
        movl TEMP, io_file
        call write_dispatch
        movl $msg_undefined_var, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_procedure: 
        movl $msg_expected_procedure, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_number: 
        movl $msg_expected_number, TEMP
        jmp error_msg

get_last_number:  #Moves first argument in ARGL into VAL and guarantees that it
                  #is a NUMBER and that it is the last argument.
                  #Type header is returned in TEMP
        #The following four lines are equivalent with:
        #call insure_one_last_arg
        testl ARGL, ARGL
        jz error_too_few_args
        cmpl $0, 4(ARGL)
        jne error_too_many_args
    get_number:
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_number
        movl (VAL), TEMP
        cmpb $NUMBER, TEMP
        jne error_expected_number
        ret
##------------------------------------------------------------------------------
error_expected_exact_integer: 
        movl $msg_expected_exact_integer, TEMP
        jmp error_msg

get_last_exact_integer: 
                #Moves first argument in ARGL into TEMP and guarantees that it
                #is an exact INTEGER and that it is the last argument.
        #The following four lines are equivalent with:
        #call insure_one_last_arg
        testl ARGL, ARGL
        jz error_too_few_args
        cmpl $0, 4(ARGL)
        jne error_too_many_args
get_exact_integer: 
        movl (ARGL), TEMP
        jecxz error_expected_integer
        cmpl $INTEGER, (TEMP)
        jne error_expected_exact_integer
        ret
##------------------------------------------------------------------------------
error_expected_integer: 
        movl $msg_expected_integer, TEMP
        jmp error_msg

get_integer_exactness: 
        #Moves first argument in ARGL into EXP and guarantees that it
        #is an INTEGER.
        #The INTEGER type header in TEMP OR (FREE) is stored in (FREE)
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_integer
        movl (EXP), TEMP
        cmpw $INTEGER, TEMP
        jne error_expected_integer
        orl TEMP, (FREE)
        ret
##------------------------------------------------------------------------------
error_expected_rational: 
        movl $msg_expected_rational, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_char: 
        movl $msg_expected_char, TEMP
        jmp error_msg

get_last_char_ascii:
                #Moves first argument in ARGL into VAL and guarantees that it
                #is a CHAR.  Then the ascii is moved into TEMP
        #The following four lines are equivalent with:
        #call insure_one_last_arg
        testl ARGL, ARGL
        jz error_too_few_args
        cmpl $0, 4(ARGL)
        jne error_too_many_args
    get_char_ascii:
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_char
        movl (VAL), VAL
        cmpb $CHAR, VAL
        jne error_expected_char
        movzbl VALH, TEMP
        ret
##------------------------------------------------------------------------------
error_expected_string: 
        movl $msg_expected_string, TEMP
        jmp error_msg

get_last_string: #Moves last argument in ARGL into ARGL and insures that it is a
                # STRING.  TEMP holds STRING header with length in high word.
        call insure_one_last_arg
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz error_expected_string
        movl (ARGL), TEMP
        cmpb $STRING, TEMP
        jne error_expected_string
        ret
##------------------------------------------------------------------------------
error_expected_mutable_string: 
        movl $msg_expected_mutable_string, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_vector: 
        movl $msg_expected_vector, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_symbol: 
        movl $msg_expected_symbol, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_pair: 
        movl $msg_expected_pair, TEMP
        jmp error_msg

insure_pair:
        testl ARGL, ARGL
        jz error_expected_pair
        testl $1, (ARGL)
        jnz error_expected_pair
        ret
##------------------------------------------------------------------------------
error_expected_input_port: 
        movl $msg_expected_input_port, TEMP
        jmp error_msg

get_last_input_port: #Moves first argument in ARGL into ARGL and guarantees that
                     #it is an INPUT_PORT.  If no argument is given, then
                     #current_input_port is used.
                     #Then sets io_file accordingly.
        testl ARGL, ARGL
        jz set_input_port_current
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz error_expected_input_port
        movl (ARGL), TEMP
        cmpb $INPUT_PORT, TEMP
        jne error_expected_input_port
        jmp set_input_port
set_input_port_current: 
        movl current_input_port, ARGL
set_input_port: 
        movl 4(ARGL), TEMP
        movl TEMP, io_file
        movl (ARGL), TEMP
        shrl $16, TEMP
        testb TEMP, TEMP
        jz set_input_port_getch
        movb TEMP, input
        ret
    set_input_port_getch:
        jmp getch
##------------------------------------------------------------------------------
error_expected_output_port: 
        movl $msg_expected_output_port, TEMP
        jmp error_msg

get_last_output_port:#Moves first argument in ARGL into ARGL and guarantees that
                     #it is an OUTPUT_PORT.  If no argument is given, then
                     #current_output_port is used.
                     #Then sets io_file accordingly.
        testl ARGL, ARGL
        jz set_output_port_current
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz error_expected_output_port
        cmpl $OUTPUT_PORT, (ARGL)
        jne error_expected_output_port
        jmp set_output_port
set_output_port_current: 
        movl current_output_port, ARGL
set_output_port:
        movl 4(ARGL), TEMP
        movl TEMP, io_file
        ret
##------------------------------------------------------------------------------
error_expected_simple_type: 
        movl $msg_expected_simple_type, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_expected_paren: 
        movl $msg_expected_paren, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_extra_eol: 
        movl $msg_extra_eol, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_invalid_pair: 
        movl $msg_invalid_pair, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_invalid_index: 
        movl $msg_invalid_index, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_overflow: 
        movl $msg_overflow, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_unknown_literal: 
        movl $msg_unknown_literal, TEMP
        jmp error_msg
##------------------------------------------------------------------------------
error_eof: 
        movl $msg_eof, TEMP
error_msg: 
        movl stderr, VAL
        movl VAL, io_file
        call puts
        movb $10, output
        call putch
        movl stackbase, SP
        movl $0, root
        xorl VAL, VAL      #These must be cleared in case they are
        xorl EXP, EXP      #not pointing to valid scheme objects.
        xorl UNEV, UNEV    #
        xorl ARGL, ARGL    #
        jmpl *error_continuation
##------------------------------------------------------------------------------
error_out_of_memory: 
        movl stderr, TEMP
        movl TEMP, io_file
        movl $msg_out_of_memory, TEMP
        call puts
        movb $10, output
        call putch
exit_not_ok: 
        movl $1, FREE
        jmp sys_exit
##------------------------------------------------------------------------------
##SPECIAL FORMS
##------------------------------------------------------------------------------
eval_dispatch: 
        testl EXP, EXP
        jz ev_self_eval
        testl $1, (EXP)
        jz ev_combination
        movl (EXP), TEMP
        cmpb $SYMBOL, TEMP
        je ev_variable
    ev_self_eval: 
        movl EXP, VAL
        ret
##------------------------------------------------------------------------------
ev_variable:     #Clobbers ENV and returns in VAL the value of the symbol in EXP
        movl (ENV), VAL        #VAL <- (car ENV) is ((symbol . value) ...)
    variable_binding_loop: 
        testl VAL, VAL
        jz variable_enclosing_scope
        movl (VAL), TEMP        #TEMP <- (car VAL) is (symbol . value)
        cmpl EXP, (TEMP)        #(eq? (car TEMP) EXP)
        je variable_found
        movl 4(VAL), VAL       #VAL <- (cdr VAL)
        jmp variable_binding_loop

    variable_enclosing_scope: 
        movl 4(ENV), ENV       #ENV <- (cdr ENV) is enclosing scope
        testl ENV, ENV
        jnz ev_variable
        jmp error_undefined_var

    variable_found: 
        movl 4(TEMP), VAL       #VAL <- (cdr TEMP) is value
        ret
##------------------------------------------------------------------------------
ev_combination: 
        movl 4(EXP), UNEV      #UNEV <- (cdr EXP)
        movl (EXP), VAL       #VAL <- (car EXP)
        testl VAL, VAL
        jz error_expected_procedure
        movl (VAL), TEMP
        cmpb $PROCEDURE, TEMP
        je apply_procedure
        cmpl $MEMOIZABLE_SYMBOL, TEMP
        je memoize
        movl VAL, EXP
        save ENV
        save UNEV
        call eval_dispatch
        restore UNEV
        restore ENV
        testl VAL, VAL
        jz error_expected_procedure
        jmp apply_procedure

    memoize: 
        save ENV
        save EXP
        movl VAL, EXP
        call ev_variable
        restore EXP
        restore ENV
        movl VAL, (EXP) #Memoizes VAL
apply_procedure:      #evaluates and applies arguments in UNEV
                     #to non-null procedure in VAL
        cmpl $SPECIAL_FORM, (VAL)
        je apply_builtin       #UNEV points to arguments
        xorl ARGL, ARGL
        testl UNEV, UNEV
        jz apply_dispatch      #(null? UNEV) means there are no operands
        save VAL
    apply_operand_loop: 
        save ARGL
        movl (UNEV), EXP       #EXP <- (car UNEV)
        cmpl $0, 4(UNEV)        #(null? (cdr UNEV))
        je apply_last_arg
        save ENV
        save UNEV
        call eval_dispatch
        restore UNEV
        restore ENV
        restore ARGL
        movl VAL, (FREE)
        movl ARGL, 4(FREE)
        movl FREE, ARGL         #ARGL <- (cons VAL ARGL)
        call advance_free
        movl 4(UNEV), UNEV      #UNEV <- (cdr UNEV)
        jmp apply_operand_loop

    apply_last_arg: 
        call eval_dispatch
        restore ARGL
        movl VAL, (FREE)
        movl ARGL, 4(FREE)
        movl FREE, ARGL
        call advance_free
        call reverse
        movl VAL, ARGL
        restore VAL
apply_dispatch: 
        cmpl $PRIMITIVE, (VAL)
        je apply_builtin       #ARGL points to arguments
        cmpl $CLOSURE, (VAL)
        je apply_compound
        cmpl $CONTINUATION, (VAL)
        je apply_continuation
        cmpl $FOREIGN_PROCEDURE, (VAL)
        je apply_foreign_procedure
        jmp error_expected_procedure

apply_builtin: 
        jmpl *4(VAL)

    apply_foreign_procedure:
        movl FREE, (dream_free)
        pushl ARGL
        call *4(VAL)
        ##movl %eax, VAL
        addl $4, SP
        xorl ENV, ENV
        xorl EXP, EXP
        xorl UNEV, UNEV
        xorl ARGL, ARGL
        movl (dream_free), FREE
        movl (dream_error), TEMP
        cmpl $0, TEMP
        jne foreign_error
        ret
    foreign_error:
        movl $0, (dream_error)
        jmp *TEMP

    apply_continuation: 
        call insure_one_last_arg
        movl 4(VAL), VAL       #VAL <- (root . stack)
        movl (VAL), TEMP
        movl TEMP, root
        movl 4(VAL), VAL       #VAL is stack
        movl stackbase, SP
    apply_continuation_loop: 
        testl VAL, VAL
        jz apply_continuation_end
        movl (VAL), TEMP
        pushl 4(TEMP)
        movl 4(VAL), VAL
        jmp apply_continuation_loop

    apply_continuation_end: 
        restore ENV
        movl (ARGL), VAL
        ret

    apply_compound: 
        movl 4(VAL), VAL     #VAL <- (cdr VAL) is (environment formals . body)
        movl (VAL), ENV      #ENV <- (car VAL) is environment
        movl 4(VAL), VAL     #VAL <- (cdr VAL) is (formals . body)
        movl (VAL), UNEV      #UNEV <- (car VAL) is formals
    apply_extend_environment: 
        movl ARGL, EXP
        xorl TEMP, TEMP    #Previous arg
    apply_extend_loop: 
        testl UNEV, UNEV
        jz apply_no_more_formals
        testl $1, (UNEV)
        jnz apply_extend_rest
        testl EXP, EXP
        jz error_too_few_args
        movl (UNEV), TEMP        #TEMP <- (car UNEV)
        movl TEMP, (FREE)        #(set-car! FREE TEMP)
        movl (EXP), TEMP        #TEMP <- (car EXP)
        movl TEMP, 4(FREE)       #(set-cdr! FREE TEMP)
        movl FREE, (EXP)        #(set-car! EXP FREE) is ((car UNEV) (car EXP))
        call advance_free
        movl 4(UNEV), UNEV       #UNEV <- (cdr UNEV) is next formal
        movl EXP, TEMP          #TEMP remembers this argument
        movl 4(EXP), EXP       #EXP <- (cdr EXP) is next argument
        jmp apply_extend_loop

    apply_extend_rest: 
        movl UNEV, (FREE)
        movl EXP, 4(FREE)       #(UNEV . EXP) for binding rest of args
        movl FREE, 8(FREE)
        addl $8, FREE
        movl $0, 4(FREE)
        jecxz apply_empty_args
        movl FREE, 4(TEMP)
        call advance_free
        jmp apply_extend_enclose

    apply_empty_args: 
        movl FREE, ARGL
        call advance_free
        jmp apply_extend_enclose

    apply_no_more_formals: 
        testl EXP, EXP
        jnz error_too_many_args
    apply_extend_enclose: 
        movl ARGL, (FREE)       #(set-car! TEMP ARGL)
        movl ENV, 4(FREE)      #(set-cdr! TEMP ENV)
        movl FREE, ENV         #ENV is (ARGL . ENV)
        call advance_free
        movl 4(VAL), UNEV      #UNEV <- (cdr VAL) is body
##------------------------------------------------------------------------------
ev_begin: 
        testl UNEV, UNEV
        jz error_too_few_args
    begin_loop: 
        movl (UNEV), EXP       #EXP <- (car UNEV)
        cmpl $0, 4(UNEV)        #(null? (cdr UNEV))
        je begin_last_exp
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
        movl 4(UNEV), UNEV      #UNEV <- (cdr UNEV) is next expression
        jmp begin_loop

    begin_last_exp: 
        jmp eval_dispatch
##------------------------------------------------------------------------------
ev_quote: 
        testl UNEV, UNEV
        jz error_too_few_args
        cmpl $0, 4(UNEV)
        jne error_too_many_args
        movl (UNEV), VAL       #VAL <- (car UNEV)
        ret
##------------------------------------------------------------------------------
ev_lambda: 
        testl UNEV, UNEV
        jz error_too_few_args
lambda: 
        movl ENV, (FREE)      #(set-car! FREE ENV)
        movl UNEV, 4(FREE)      #(set-cdr! FREE (formals . body))
        movl FREE, VAL
        addl $8, FREE
        movl $CLOSURE, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
ev_if: 
        testl UNEV, UNEV
        jz error_too_few_args
        movl (UNEV), EXP       #EXP <- (car UNEV) is predicate
        movl 4(UNEV), UNEV      #UNEV <- (cdr UNEV) is (consequent alternate)
        testl UNEV, UNEV
        jz error_too_few_args
        save UNEV
        save ENV
        call eval_dispatch

        restore ENV
        restore UNEV
        cmpl $false, VAL
        jne if_consequent
    if_alternate: 
        movl 4(UNEV), UNEV       #UNEV <- (cdr UNEV) is (alternate)
        testl UNEV, UNEV
        jnz if_consequent
        ret                     #No alternate
    if_consequent: 
        movl (UNEV), EXP        #EXP <- (car UNEV)
        jmp eval_dispatch
##------------------------------------------------------------------------------
ev_set: 
        testl UNEV, UNEV
        jz error_too_few_args
        movl 4(UNEV), EXP       #EXP <- (cdr UNEV) is (value)
        testl EXP, EXP
        jz error_too_few_args
        cmpl $0, 4(EXP)
        jne error_too_many_args
        movl (EXP), EXP        #EXP <- (car EXP) is value
        movl (UNEV), UNEV        #UNEV <- (car UNEV) is symbol
        save UNEV
        save ENV
        call eval_dispatch

        restore ENV
        restore UNEV
    set_scope_loop: 
        movl (ENV), TEMP     #TEMP <- (car ENV) is ((symbol . value) ...)
    set_binding_loop: 
        jecxz set_enclosing_scope
        movl (TEMP), EXP        #EXP <- (car TEMP) is (symbol . value)
        cmpl UNEV, (EXP)        #(eq? (car EXP) UNEV)
        je set_found
        movl 4(TEMP), TEMP       #TEMP <- (cdr TEMP) is next binding
        jmp set_binding_loop

    set_enclosing_scope: 
        movl 4(ENV), ENV       #ENV <- (cdr ENV) is enclosing scope
        testl ENV, ENV
        jnz set_scope_loop
        movl UNEV, EXP
        jmp error_undefined_var

    set_found: 
        movl VAL, 4(EXP)       #(set-cdr! EXP VAL)
        ret
##------------------------------------------------------------------------------
ev_define: 
        testl UNEV, UNEV
        jz error_too_few_args
        movl 4(UNEV), EXP       #EXP <- (cdr UNEV) is (value)
        testl EXP, EXP
        jz error_too_few_args
        movl (UNEV), UNEV        #UNEV <- (car UNEV) is variable to be
        testl UNEV, UNEV
        jz error_expected_symbol
        testl $1, (UNEV)
        jz define_lambda
        cmpl $0, 4(EXP)
        jne error_too_many_args
        movl (EXP), EXP        #EXP <- (car EXP) is value
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
    define_define: 
        movl (ENV), TEMP     #TEMP <- (car ENV) is ((symbol . value) ...)
    define_binding_loop: 
        jecxz define_new_var
        movl (TEMP), EXP        #EXP <- (car TEMP) is (symbol . value)
        cmpl UNEV, (EXP)
        je define_found
        movl 4(TEMP), TEMP       #TEMP <- (cdr TEMP)
        jmp define_binding_loop

    define_new_var: 
        movl UNEV, (FREE)        #(set-car! FREE UNEV)
        movl VAL, 4(FREE)       #(set-cdr! FREE VAL)
        movl FREE, 8(FREE)       #(set-car! FREE+8 FREE) is (UNEV . VAL)
        addl $8, FREE
        movl (ENV), TEMP        #TEMP <- (car ENV) is bindings
        movl TEMP, 4(FREE)       #(set-cdr! FREE TEMP)
        movl FREE, (ENV)        #(set-car ENV FREE) is ((symbol . value) ENV)
        jmp advance_free

    define_found: 
        movl VAL, 4(EXP)       #(set-cdr! EXP VAL)
        ret

    define_lambda: 
        movl 4(UNEV), VAL
        movl VAL, (FREE)        #formals
        movl EXP, 4(FREE)       #body
        movl (UNEV), EXP        #remember symbol for later
        movl FREE, UNEV
        call advance_free
        call lambda
        movl EXP, UNEV
        jmp define_define
##------------------------------------------------------------------------------
ev_and: 
        testl UNEV, UNEV
        jz return_true
    ev_and_loop: 
        movl (UNEV), EXP
        movl 4(UNEV), UNEV
        testl UNEV, UNEV
        jz eval_dispatch
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
        cmpl $false, VAL
        jne ev_and_loop
        ret
##------------------------------------------------------------------------------
ev_or: 
        testl UNEV, UNEV
        jz return_false
    ev_or_loop: 
        movl (UNEV), EXP
        movl 4(UNEV), UNEV
        testl UNEV, UNEV
        jz eval_dispatch
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
        cmpl $false, VAL
        je ev_or_loop
        ret
##------------------------------------------------------------------------------
    ev_cond_loop: 
        movl 4(UNEV), UNEV
ev_cond: 
        testl UNEV, UNEV
        jz return_false
        movl (UNEV), ARGL
        call insure_pair
        movl (ARGL), EXP
        cmpl $sym_else, EXP
        je ev_cond_true
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
        cmpl $false, VAL
        je ev_cond_loop
        movl (UNEV), ARGL #This is cheaper than saving and retrieving it
    ev_cond_true: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz ev_cond_end
        testl $1, (ARGL)
        jnz error_expected_pair
        movl (ARGL), EXP
        cmpl $sym_arrow, EXP
        jne ev_cond_begin
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        movl VAL, (FREE)
        movl $0, 4(FREE)
        movl FREE, UNEV
        movl (ARGL), EXP
        call advance_free
        save UNEV
        save ENV
        call eval_dispatch
        restore ENV
        restore UNEV
        jmp prim_apply_direct

    ev_cond_begin:
        movl ARGL, UNEV
        jmp ev_begin

    ev_cond_end:
        ret
##------------------------------------------------------------------------------
ev_case: 
        testl UNEV, UNEV
        jz error_too_few_args
        movl (UNEV), EXP
        save ENV
        save UNEV
        call eval_dispatch
        restore UNEV
        restore ENV
    ev_case_clause_loop: 
        movl 4(UNEV), UNEV
        testl UNEV, UNEV
        jz return_false
        movl (UNEV), ARGL
        call insure_pair
        movl (ARGL), ARGL
        cmpl $sym_else, ARGL
        je ev_case_found
    ev_case_memv_loop: 
        testl ARGL, ARGL
        jz ev_case_clause_loop
        testl $1, ARGL
        jnz error_expected_pair
        movl (ARGL), EXP
        movl 4(ARGL), ARGL
        cmpl EXP, VAL
        je ev_case_found
        testl VAL, VAL
        jz ev_case_memv_loop
        testl EXP, EXP
        jz ev_case_memv_loop
        movl (VAL), TEMP
        cmpl (EXP), TEMP
        jne ev_case_memv_loop
        testl $1, TEMP        #Pairs are not allowed to be eqv? unless eq?
        jz ev_case_memv_loop
        cmpw $VECTOR, TEMP    #Vectors are not allowed to be eqv? unless eq?
        je ev_case_memv_loop
        movl 4(VAL), TEMP
        cmpl 4(EXP), TEMP
        jne ev_case_memv_loop
    ev_case_found: 
        movl (UNEV), ARGL #restore this clause
        movl 4(ARGL), UNEV
        jmp ev_begin
##------------------------------------------------------------------------------
let_setup: #Sets ARGL <- binding list and saves body, or jumps directly to begin
        testl UNEV, UNEV
        jz error_too_few_args
        movl (UNEV), ARGL        #should be ((formal value) ...)
        movl 4(UNEV), UNEV       #should be (body1 body2 ...)
        testl ARGL, ARGL
        jz let_setup_end
        testl $1, (ARGL)
        jnz error_expected_pair
    let_setup_end: 
        save UNEV
        ret

ev_let: 
        call let_setup
        xorl VAL, VAL          #will be bindings in new environment
        jmp ev_let_loop_begin

    ev_let_loop: 
        call let_bind
    ev_let_loop_begin: 
        testl ARGL, ARGL
        jnz ev_let_loop

        movl VAL, (FREE)
        movl ENV, 4(FREE)
        movl FREE, ENV    #should be (((formal . value) ...) . enclosing_scope)
        call advance_free
        restore UNEV
        jmp ev_begin

let_bind: 
        save ARGL
        movl (ARGL), ARGL        #should be (formal value)
        call insure_pair
        movl 4(ARGL), ARGL       #should be (value)
        call insure_one_last_arg
        movl (ARGL), EXP        #should be value
        save VAL
        save ENV
        call eval_dispatch
        restore ENV
        movl VAL, 4(FREE)
        restore VAL
        restore ARGL
        movl (ARGL), EXP        #should be (formal value) again
        movl (EXP), EXP        #should be formal
        movl EXP, (FREE)
        movl FREE, EXP
        call advance_free
        movl EXP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        call advance_free
        movl 4(ARGL), ARGL       #should be next ((formal value) ...)
        ret
##------------------------------------------------------------------------------
ev_letrec: 
        call let_setup
        xorl VAL, VAL          #will be bindings in new environment
        movl VAL, (FREE)
        movl ENV, 4(FREE)
        movl FREE, ENV    #should be (((formal . value) ...) . enclosing_scope)
        call advance_free
        jmp ev_letrec_loop_begin

    ev_letrec_loop: 
        call let_bind
    ev_letrec_loop_begin: 
        testl ARGL, ARGL
        jnz ev_letrec_loop

        movl VAL, (ENV)
        restore UNEV
        jmp ev_begin
##------------------------------------------------------------------------------
ev_let_nested: 
        call let_setup
    ev_let_nested_loop: 
        xorl VAL, VAL          #will be bindings in new environment
        movl VAL, (FREE)
        movl ENV, 4(FREE)
        movl FREE, ENV    #should be (((formal . value) ...) . enclosing_scope)
        call advance_free
        testl ARGL, ARGL
        jz ev_let_nested_end
        call let_bind
        movl VAL, (ENV)
        testl ARGL, ARGL
        jnz ev_let_nested_loop

    ev_let_nested_end: 
        restore UNEV
        jmp ev_begin
##------------------------------------------------------------------------------
ev_quasiquote:
        pushl $0        #Nesting Depth
        call ev_quasiquote_nested
        addl $4, SP
        ret

    ev_quasiquote_nested:
        incl 4(SP)
        testl UNEV, UNEV
        jz error_too_few_args
        cmpl $0, 4(UNEV)
        jnz error_too_many_args
        movl (UNEV), EXP
    ev_quasiquote_dispatch:
        testl EXP, EXP
        jz ev_quasiquote_simple
        testl $1, (EXP)
        jz ev_quasiquote_pair
        movl (EXP), TEMP
        cmpw $VECTOR, TEMP
        je ev_quasiquote_vector
    ev_quasiquote_simple:
        movl EXP, VAL
        ret

    ev_quasiquote_car_splice:
        save VAL       #New car
        movl 4(EXP), EXP      #Old cdr
        pushl 4(SP)
        call ev_quasiquote_dispatch
        addl $4, SP
        restore EXP    #New car
        jmp append

    ev_quasiquote_pair:
        movl (EXP), VAL
        cmpl $sym_unquote, VAL
        je ev_unquote_template
        cmpl $sym_unquote_splicing, VAL
        je ev_unquote_splicing_template
        cmpl $sym_quasiquote, VAL
        je ev_quasiquote_template
    ev_quasiquote_car:
        save EXP
        movl VAL, EXP
        pushl 4(SP)
        call ev_quasiquote_dispatch
        restore EXP    #Original pair
        addl $4, SP
        cmpl $0, -4(SP) #If it became negative, then we must splice.
        js ev_quasiquote_car_splice
    ev_quasiquote_car_simple:
        save VAL       #New car
        movl 4(EXP), EXP      #Old cdr
        pushl 4(SP)
        call ev_quasiquote_dispatch
        addl $4, SP
        restore EXP    #New car
    ev_quasiquote_cons:
        movl EXP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free

    ev_unquote_splicing_template:
        decl 4(SP)
        jnz ev_quasiquote_car_simple
        decl 4(SP) #This will inform ev_quasiquote_car that this is a splice.
        jmp ev_unquote
        
    ev_unquote_template:
        decl 4(SP)
        jnz ev_quasiquote_car_simple
    ev_unquote:
        movl 4(EXP), EXP
        testl EXP, EXP
        jz error_too_few_args
        cmpl $0, 4(EXP)
        jnz error_too_many_args
        movl (EXP), EXP
        jmp eval_dispatch

    ev_quasiquote_template:
        incl 4(SP)
        jmp ev_quasiquote_car_simple

    ev_quasiquote_vector:
        call vector_to_list
        movl VAL, EXP
        pushl 4(SP)
        call ev_quasiquote_dispatch
        addl $4, SP
        jmp list_to_vector
##------------------------------------------------------------------------------
##APPLY
##------------------------------------------------------------------------------
prim_apply: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_procedure
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz error_too_few_args
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        movl (ARGL), UNEV
        testl UNEV, UNEV
        jz prim_apply_direct
        testl $1, (UNEV)
        jnz error_expected_pair
prim_apply_direct:  #Applies non-null procedure in VAL to args in UNEV
                   #WITHOUT evaluating them.
        movl (VAL), TEMP
        cmpb $PROCEDURE, TEMP
        jne error_expected_procedure
        cmpw $SPECIAL_FORM, TEMP
        je apply_builtin
        movl UNEV, ARGL
        jmp apply_dispatch
##------------------------------------------------------------------------------
##INPUT/OUTPUT PRIMITIVES
##------------------------------------------------------------------------------
prim_read_char: 
        call get_last_input_port
        movl $INPUT_PORT, (ARGL) #Clear peek ascii
        movzbw input, TEMPW
        jecxz read_eof
        leal chararray(, TEMP, 4), VAL
        ret
##------------------------------------------------------------------------------
prim_peek_char: 
        call get_last_input_port
        movzbw input, TEMPW
        shll $16, TEMP
        movb $INPUT_PORT, TEMP
        movl TEMP, (ARGL) #Remember peeked ascii
        movzbl input, TEMP
        jecxz read_eof
        lea chararray(, TEMP, 4), VAL
        ret
##------------------------------------------------------------------------------
    read_eof: 
        movl $eof, VAL
        ret
##------------------------------------------------------------------------------
prim_read: 
        call get_last_input_port
prim_read_argl: 
        save ARGL       #Remember port so that we can update peek char
        call read_dispatch
        restore ARGL
        movzbl input, TEMP
        shll $16, TEMP
        movb $INPUT_PORT, TEMP
        movl TEMP, (ARGL)
        cmpl $sym_eol, VAL
        je error_extra_eol
        ret

    read_input: 
        call getch
read_dispatch: 
        cmpb $0, input
        je read_eof
        cmpb $' ', input
        je read_input
        cmpb $9, input           #tab
        je read_input
        cmpb $10, input          #newline
        je read_input
        cmpb $'(', input
        je read_list
        cmpb $')', input
        je read_eol
        cmpb $'#', input
        je read_literal
        cmpb $0x27, input        #apostrophe
        je read_quote
        cmpb $'`', input
        je read_quasiquote
        cmpb $',', input
        je read_unquote
        cmpb $'.', input
        je read_dot
        cmpb $0x22, input        #double quote
        je read_string
        cmpb $';', input
        je read_comment
        cmpb $'0', input
        jb read_symbol
        cmpb $'9', input
        ja read_symbol
        jmp read_exact_decimal_number

    read_dot: 
        call getch
        movb input, VAL
        cmpb $' ', VAL
        je read_dot_confirmed
        cmpb $9, VAL
        je read_dot_confirmed
        cmpb $10, VAL
        je read_dot_confirmed
        cmpb $'(', VAL
        je read_dot_confirmed
        cmpb $')', VAL
        je read_dot_confirmed
        cmpb $';', VAL
        je read_dot_confirmed
        cmpb $0, VAL
        je read_dot_confirmed
    read_dot_prefix: 
        movl freesymbol, ARGL
        cld
        movl $SYMBOL, (ARGL)
        movb $'.', 4(ARGL)
        addl $5, ARGL
        jmp read_symbol_chars

    read_dot_confirmed: 
        movl $sym_dot, VAL
        ret

    read_comment: 
        call getch
        cmpb $0, input
        je error_eof
        cmpb $10, input          #newline
        jne read_comment
        jmp read_input

    read_eol: 
        call getch
        movl $sym_eol, VAL
        ret

    read_literal: 
        call getch
        cmpb $0, input
        je error_eof
        cmpb $'(', input
        je read_vector
        movb input, TEMP
        to_lower TEMP
        call getch
        cmpb $'t', TEMP
        je read_boolean_true
        cmpb $'f', TEMP
        je read_boolean_false
        cmpb $0x5c, TEMP     #backslash
        je read_character
        cmpb $'!', TEMP
        je read_comment
        movl $INTEGER, (FREE)
        cmpb $'e', TEMP
        je read_exact_prefix
        cmpb $'i', TEMP
        je read_inexact_prefix
    read_number_literal:
        cmpb $'x', TEMP
        je read_hexadecimal_number
        cmpb $'o', TEMP
        je read_octal_number
        cmpb $'b', TEMP
        je read_binary_number
        cmpb $'d', TEMP
        je read_decimal_number
        jmp error_unknown_literal

    read_inexact_prefix:
        orl $0x10000, (FREE)
    read_exact_prefix:
        cmpb $0, input
        je error_eof
        cmpb $'#', input
        jne read_decimal_number
        call getch
        movb input, TEMP
        cmpb $0, TEMP
        je error_eof
        to_lower TEMP
        call getch
        jmp read_number_literal

    read_boolean_true: 
        movl $true, VAL
        ret

    read_boolean_false: 
        movl $false, VAL
        ret

    read_character: 
        cmp $' ', input
        je read_character_direct
        cmpb $10, input
        je read_character_direct
        cmpb $9, input
        je read_character_direct
        cmpb $'(', input
        je read_character_direct
        cmpb $')', input
        je read_character_direct
        cmpb $';', input
        je read_character_direct
        movb input, TEMP
        movb TEMP, backup_char
        call read_symbol
        movb 5(VAL), TEMP
        testb TEMP, TEMP
        jz read_character_simple
        cmpl $sym_space_name, VAL
        je read_character_space
        cmpl $sym_newline, VAL
        je read_character_newline
        jmp error_unknown_literal

    read_character_direct: 
        movzbl input, VAL
        leal chararray(, VAL, 4), VAL
        jmp getch

    read_character_simple: 
        movzbl backup_char, VAL
        leal chararray(, VAL, 4), VAL
        ret

    read_character_simple_space: 
        call getch
    read_character_space: 
        leal 32*4+chararray, VAL
        ret

    read_character_simple_newline: 
        call getch
    read_character_newline: 
        leal 40+chararray, VAL
        ret

    read_binary_number: 
        movl $0b10, radix
        jmp read_number

    read_octal_number: 
        movl $010, radix
        jmp read_number

    read_exact_decimal_number:
        movl $INTEGER, (FREE)
    read_decimal_number: 
        movl $10, radix
        jmp read_number

    read_hexadecimal_number: 
        movl $0x10, radix
        jmp read_number

    read_quasiquote: 
        pushl $sym_quasiquote
        jmp read_abbreviation

    read_unquote: 
        call getch
        cmpb $'@', input
        je read_unquote_splicing
        pushl $sym_unquote
        jmp read_abbreviation_contents

    read_unquote_splicing: 
        pushl $sym_unquote_splicing
        jmp read_abbreviation

    read_quote: 
        pushl $sym_quote
    read_abbreviation: 
        call getch
    read_abbreviation_contents: 
        call read_dispatch
        cmpl $sym_eol, VAL
        je error_extra_eol
        movl VAL, (FREE)
        movl $0, 4(FREE)             #(cons VAL ())
        movl FREE, VAL
        addl $8, FREE
        movl VAL, 4(FREE)
        popl (FREE)
        movl FREE, VAL
        jmp advance_free

    read_list: 
        save ENV
        xorl ARGL, ARGL
        call getch
    read_list_loop: 
        save ARGL
        call read_dispatch
        restore ARGL
        cmpl $sym_eol, VAL
        je read_list_end
        cmpl $eof, VAL
        je error_eof
        cmpl $sym_dot, VAL
        je read_cdr
        movl VAL, (FREE)
        movl $0, 4(FREE)             #(cons VAL ())
        movl FREE, VAL
        call advance_free
        testl ARGL, ARGL
        jz read_list_first
        movl VAL, 4(ENV)
        movl VAL, ENV
        jmp read_list_loop

    read_list_first: 
        movl VAL, ARGL
        movl VAL, ENV
        jmp read_list_loop

    read_list_end: 
        movl ARGL, VAL
        restore ENV
        ret

    read_cdr: 
        testl ARGL, ARGL
        jz error_invalid_pair
        save ARGL
        call read_dispatch
        restore ARGL
        cmpl $sym_eol, VAL
        je error_invalid_pair
        cmpl $eof, VAL
        je error_eof
        cmpl $sym_dot, VAL
        je error_invalid_pair
        movl VAL, 4(ENV)
        save ARGL
        call read_dispatch
        cmpl $sym_eol, VAL
        jne error_expected_paren
        restore VAL
        jmp read_list_end

    read_symbol: 
        movl freesymbol, ARGL
        cld
        movl $SYMBOL, (ARGL)
        addl $4, ARGL
        movb input, VAL
    read_symbol_chars: 
        to_lower VAL
        stosb
        call getch
        cmpb $'-', VAL
        jne read_name_loop
        cmpb $'0', input
        jb read_name_loop
        cmpb $'9', input
        ja read_name_loop
        xorl VAL, VAL          #Clear scheme register
        xorl ARGL, ARGL          #Clear scheme register
        call read_exact_decimal_number
        cmpl $INTEGER, (VAL)
        jne read_negate_rational
        negl 4(VAL)
        jo error_overflow
        ret
    read_negate_rational:
        movl 4(VAL), TEMP
        movl (TEMP), TEMP
        negl 4(TEMP)
        jo error_overflow
        ret

    read_name_loop: 
        movb input, VAL
        cmpb $0, VAL
        je read_name_end
        cmpb $' ', VAL
        je read_name_end
        cmpb $10, VAL          #newline
        je read_name_end
        cmpb $9, VAL           #tab
        je read_name_end
        cmpb $'(', VAL
        je read_name_end
        cmpb $')', VAL
        je read_name_end
        cmpb $';', VAL
        je read_name_end
        to_lower VAL
        stosb
        call getch
        jmp read_name_loop

    read_name_end: 
        movb $0, (ARGL)
        incl ARGL
        testl $3, ARGL            #Are we at 4 byte boundary?
        jnz read_name_end
        movl ARGL, backup       #new freesymbol in case this symbol is new
        movl freesymbol, VAL    #symbol we are interning
        movl 4(VAL), TEMP       #first letter
        andl $0x7f, TEMP
        shll $HASHPOWER, TEMP
        addl $obhash, TEMP
                     #TEMP is now indexed into obhash according to first letter
    read_compare_loop: 
        cmpl $0, 0(TEMP)      #No more symbols?
        je read_new_symbol
        movl 0(TEMP), UNEV         #UNEV is symbol
        movl freesymbol, ARGL     #ARGL is read symbol
        cmpsl                    #Skip SYMBOL type header
    read_symbol_compare_loop: 
        cmpsb
        jne read_next_symbol
        cmpb $0, -1(UNEV)        #end of read symbol?
        jne read_symbol_compare_loop
        movl 0(TEMP), VAL        #Found! VAL is matching symbol
        xorl UNEV, UNEV      #Clear scheme register
        xorl ARGL, ARGL      #Clear scheme register
        ret

    read_next_symbol: 
        addl $4, TEMP
        cmpl $obhash_end, TEMP
        jne read_compare_loop
        movl $obhash, TEMP
        jmp read_compare_loop

    read_new_symbol: 
        movl backup, ARGL        #ARGL is next free symbol space
        movl ARGL, freesymbol
        movl VAL, 0(TEMP)        #set to new symbol
        xorl UNEV, UNEV      #Clear scheme register
        xorl ARGL, ARGL      #Clear scheme register
        ret

    read_rational:
        call getch
        call digitize_stream
        testl EXP, EXP
        jnz error_unknown_literal
        movl 4(FREE), TEMP
        jmp make_rational

    read_number: 
        movl $getch, thunk
        call digitize_stream_nonempty
        testl EXP, EXP
        jnz error_unknown_literal
        movl VAL, 4(FREE)
        movl FREE, VAL
        cmpb $'/', input
        je read_rational
        jmp advance_free

    read_string: 
        movl freestring, ARGL
        movl ARGL, 4(FREE)       #We will make the STRING type later
        cld
    read_string_loop: 
        call getch
        movb input, TEMP
        cmpb $0x22, TEMP          #double quote
        je read_string_end
        cmpb $'\\', TEMP
        jne read_string_continue
        call getch
        movb input, TEMP
    read_escape_newline: 
        cmpb $'n', TEMP
        jne read_escape_tab
        movb $10, TEMP            #newline
    read_escape_tab: 
        cmpb $'t', TEMP
        jne read_string_continue
        movb $9, TEMP             #tab
    read_string_continue: 
        testb TEMP, TEMP
        jz error_eof
        movb TEMP, (ARGL)
        incl ARGL
        cmpl memstrlimit, ARGL
        jae read_string_gc
        jmp read_string_loop

    read_string_end: 
        movb $0, (ARGL)
        movl ARGL, TEMP
        subl freestring, TEMP
        shll $16, TEMP
        movw $MUTABLE_STRING, TEMP
        movl TEMP, (FREE)
        incl ARGL
        movl ARGL, freestring
        xorl ARGL, ARGL      #Clear scheme register
        movl FREE, VAL
        call advance_free
        jmp getch

    read_string_gc: 
        movb $0, (ARGL)      #End the string temporarily.
        pushl freestring
        xorl ARGL, ARGL      #Clear scheme register
        call gc_strings
        popl UNEV
        movl freestring, ARGL
        movl ARGL, 4(FREE)   #New string pointer
        cld
        jmp read_string_gc_recover
    read_string_gc_loop: 
        stosb
    read_string_gc_recover: 
        lodsb
        testb VAL, VAL
        jnz read_string_gc_loop
    read_string_gc_end: 
        xorl UNEV, UNEV      #Clear scheme register
        xorl VAL, VAL      #Clear scheme register
        cmpl memstrlimit, ARGL
        jae error_out_of_memory
        jmp read_string_loop

    read_vector: 
        call read_list
        jmp list_to_vector
##------------------------------------------------------------------------------
prim_write_char: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_char
        movl 4(ARGL), ARGL
        call get_last_output_port
        movl (VAL), TEMP
        cmpb $CHAR, TEMP
        jne error_expected_char
        movb TEMPH, output
        jmp putch
##------------------------------------------------------------------------------
prim_newline:
        call get_last_output_port
        movb $10, output
        jmp putch
##------------------------------------------------------------------------------
prim_display: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL      #VAL <- (car ARGL) is object to display
        movl 4(ARGL), ARGL
        call get_last_output_port
        movl $false, UNEV     #Signals to display chars and strings
        jmp display_dispatch
##------------------------------------------------------------------------------
prim_write: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL      #VAL <- (car ARGL) is object to write
        movl 4(ARGL), ARGL
        call get_last_output_port
write_dispatch: 
        movl $true, UNEV  #Signals to write chars and strings as scheme
display_dispatch: 
        testl VAL, VAL
        jz write_empty_list
        movl (VAL), TEMP
        testl $1, TEMP
        jz write_pair
        cmpb $NUMBER, TEMP
        je write_number
        cmpb $SYMBOL, TEMP
        je write_symbol
        cmpb $CHAR, TEMP
        je write_char
        cmpb $STRING, TEMP
        je write_string
        movb $'#', output
        call putch
        cmpl $true, VAL
        je write_boolean_true
        cmpl $false, VAL
        je write_boolean_false
        cmpw $VECTOR, TEMP
        je write_vector
        movb $'<', output
        call putch
        cmpw $CLOSURE, TEMP
        je write_closure
        cmpw $CONTINUATION, TEMP
        je write_continuation
        cmpl $eof, VAL
        je write_eof
        cmpw $PRIMITIVE, TEMP
        je write_primitive
        cmpw $SPECIAL_FORM, TEMP
        je write_special_form
        jmp write_other

    write_inexact_prefix:
        movb $'#', output
        call putch
        movb $'i', output
        call putch
        jmp write_number_value

    write_number:
        testl $0x10000, TEMP
        jnz write_inexact_prefix
    write_number_value:
        cmpw $INTEGER, TEMP
        je write_decimal_number
        movl 4(VAL), VAL
        cmpw $RATIONAL, TEMP
        je write_rational_number
        jmp write_pair

    write_rational_number:
        pushl VAL
        movl (VAL), VAL
        call write_decimal_number
        movb $'/', output
        call putch
        popl VAL
        movl 4(VAL), VAL
        jmp write_decimal_number
        
    write_empty_list: 
        movb $'(', output
        call putch
        movb $')', output
        jmp putch

    write_boolean_true: 
        movb $'t', output
        jmp putch

    write_boolean_false: 
        movb $'f', output
        jmp putch

    write_char: 
        cmpl $true, UNEV
        jne display_char
        movb $'#', output
        call putch
        movb $0x5c, output       #backslash
        call putch
    display_char: 
        movb TEMPH, output
        jmp putch

    write_string: 
        cmpl $true, UNEV
        jne display_string
        movb $0x22, output       #double quote
        call putch
        movl 4(VAL), TEMP
        call puts_escape
        movb $0x22, output       #double quote
        jmp putch

    display_string: 
        movl 4(VAL), TEMP
        jmp puts

    write_symbol: 
        movl VAL, TEMP
        addl $4, TEMP
        jmp puts

    write_pair: 
        movb $'(', output
        call putch
    write_car: 
        pushl VAL
        movl (VAL), VAL      #VAL <- (car VAL)
        call display_dispatch
    write_cdr: 
        movb $' ', output
        call putch
        popl VAL
        movl 4(VAL), VAL     #VAL <- (cdr VAL)
        testl VAL, VAL
        jz write_end_pair
        testl $1, (VAL)
        jz write_car
        movb $'.', output
        call putch
        movb $' ', output
        call putch
        call display_dispatch
    write_end_pair: 
        movb $')', output
        jmp putch

    write_hexadecimal_number: 
        movl $16, radix
        jmp write_digits

    write_decimal_number: 
        movl $10, radix
    write_digits: 
        movl 4(VAL), VAL
        call stringify_number
        jmp puts

    write_closure: 
        movl $msg_closure, TEMP
        call puts
        movl 4(VAL), VAL     #VAL is (environment formals . body)
        movl 4(VAL), VAL     #VAL <- (cdr VAL) is (formals . body)
        movl (VAL), VAL      #VAL <- (car VAL) is formals
        call display_dispatch
        jmp write_close_bracket

    write_continuation: 
        movl $msg_continuation, TEMP
        call puts
        movl 4(VAL), VAL     #VAL is (root . stack)
        movl 4(VAL), VAL     #VAL is stack
        call display_dispatch
    write_close_bracket: 
        movb $'>', output
        jmp putch

    write_primitive: 
        movl $msg_primitive, TEMP
        call puts
        jmp write_other

    write_special_form: 
        movl $msg_special, TEMP
        call puts
    write_other: 
        pushl $write_close_bracket
        jmp write_hexadecimal_number

    write_eof: 
        movl $ext_eof, TEMP
        jmp puts

    write_vector: 
        movb $'(', output
        call putch
        shrl $16, TEMP   #TEMP now holds length of vector
        jmp write_vector_loop_begin
    write_vector_loop: 
        addl $4, VAL
        pushl VAL
        pushl TEMP
        movl (VAL), VAL
        call display_dispatch
        movb $' ', output
        call putch
        popl TEMP
        popl VAL
    write_vector_loop_begin: 
        decl TEMP
        jnl write_vector_loop
    write_end_vector: 
        movb $')', output
        call putch
        xorl VAL, VAL  #Clear scheme register
        ret
##------------------------------------------------------------------------------
prim_current_input_port:
        mov current_input_port, VAL
        ret
##------------------------------------------------------------------------------
prim_current_output_port:
        mov current_output_port, VAL
        ret
##------------------------------------------------------------------------------
prim_open_output_file: 
        movl $OUTPUT_PORT, (FREE)
        movl $O_WRCRTR, io_file
        jmp open_io_file
##------------------------------------------------------------------------------
prim_open_input_file: 
        movl $INPUT_PORT, (FREE)
        movl $O_RDONLY, io_file
open_io_file: 
        call get_last_string
        movl 4(ARGL), TEMP
open_io_file_scratch: 
        call open_file
        movl io_file, TEMP
        movl TEMP, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_load: 
        call get_last_string
        movl 4(ARGL), TEMP
prim_load_scratch: 
        movl $O_RDONLY, io_file
        movl $INPUT_PORT, (FREE)
        call open_io_file_scratch
        movl VAL, ARGL
        call load_scheme_argl
        movl 4(ARGL), TEMP
        jmp close_file

load_scheme_argl:
        save ARGL
        jmp load_scheme_loop_begin

    load_scheme_loop: 
        movl VAL, EXP
        movl global_environment, ENV
        call eval_dispatch
        movl root, ARGL  #Peek
        movl (ARGL), ARGL #ARGL is Input port
    load_scheme_loop_begin: 
        call set_input_port
        call prim_read_argl
        cmpl $eof, VAL
        jne load_scheme_loop
        restore ARGL
        ret
##------------------------------------------------------------------------------
prim_close_input_port: 
        call insure_one_last_arg
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz error_expected_input_port
        movl (ARGL), TEMP
        cmpb $INPUT_PORT, TEMP
        jne error_expected_input_port
        movl 4(ARGL), TEMP
        movl $-1, 4(ARGL)
        jmp close_file
##------------------------------------------------------------------------------
prim_close_output_port: 
        call insure_one_last_arg
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz error_expected_output_port
        movl (ARGL), TEMP
        cmpb $OUTPUT_PORT, TEMP
        jne error_expected_output_port
        movl 4(ARGL), TEMP
        movl $-1, 4(ARGL)
        jmp close_file
##------------------------------------------------------------------------------
prim_exit: 
        xorl FREE, FREE
        testl ARGL, ARGL
        jz sys_exit
        cmpl $0, 4(ARGL)
        jne error_too_many_args
        call get_exact_integer
        movl 4(TEMP), FREE
        jmp sys_exit
##------------------------------------------------------------------------------
##MATHEMATICAL PRIMITIVES
##------------------------------------------------------------------------------
return_with_given_header:
        movl VAL, 4(FREE)
        movl FREE, VAL
        xorl EXP, EXP
        jmp advance_free
##------------------------------------------------------------------------------
prim_add: 
        movl $INTEGER, (FREE)
        xorl VAL, VAL
    add_loop: 
        testl ARGL, ARGL
        jz return_with_given_header
        call get_integer_exactness
        addl 4(EXP), VAL
        jo error_overflow
        movl 4(ARGL), ARGL     #ARGL <- (cdr ARGL)
        jmp add_loop
##------------------------------------------------------------------------------
prim_sub: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl 4(EXP), VAL
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz additive_inverse
    sub_loop: 
        call get_integer_exactness
        subl 4(EXP), VAL
        jo error_overflow
        movl 4(ARGL), ARGL     #ARGL <- (cdr ARGL)
        testl ARGL, ARGL
        jnz sub_loop
        jmp return_with_given_header
    additive_inverse: 
        negl VAL
        jo error_overflow
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_mult: 
        movl $INTEGER, (FREE)
        movl $1, VAL
        #cltd
    mult_loop: 
        testl ARGL, ARGL
        jz return_with_given_header
        call get_integer_exactness
        imull 4(EXP)
        jc error_overflow
        movl 4(ARGL), ARGL      #ARGL <- (cdr ARGL)
        testl VAL, VAL
        jz mult_zero
        jmp mult_loop

    mult_zero:            #We have multiplied by a zero with header TEMP
        andl TEMP, (FREE) #therefore the result can be as exact as the zero was.
        jmp mult_loop
##------------------------------------------------------------------------------
prim_div: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl 4(EXP), VAL
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz multiplicative_inverse
        xorl EXP, EXP
    div_loop: 
        call get_integer_exactness
        movl 4(EXP), TEMP
        movl VAL, UNEV     #In case we need it to form a rational.
        cltd
        idivl TEMP
        testl EXP, EXP
        movl 4(ARGL), ARGL      #ARGL <- (cdr ARGL)
        jnz make_rational_last
        testl ARGL, ARGL
        jnz div_loop
        xorl UNEV, UNEV
        jmp return_with_given_header

    multiplicative_inverse: 
        cmpl $1, VAL    #A simple case of 1/1=1
        je return_exp
        cmpl $-1, VAL   #A simple case of 1/-1=-1
        je return_exp
        movl $1, TEMP
        jmp make_rational
    return_exp:
        mov EXP, VAL
        ret

    make_rational_last:
        testl ARGL, ARGL
        jnz error_expected_integer
        movl TEMP, VAL
        movl UNEV, TEMP
make_rational: #Numerator is TEMP, Denominator is VAL, (FREE) shows exactness
        testl VAL, VAL
        jz error_overflow
        jg make_normal_rational
        negl VAL        #Denominator is negative, so let's negate it and
        jo error_overflow
        negl TEMP       #the numerator.
        jo error_overflow
    make_normal_rational:
        pushl VAL
        pushl TEMP
        testl TEMP, TEMP
        jg rational_gcd_loop_begin
        negl TEMP       #Use absolute values for gcd calculation.
        jo error_overflow
        jmp rational_gcd_loop_begin

    rational_integer:
        movl TEMP, 4(FREE)
        movl FREE, VAL
        xorl UNEV, UNEV
        jmp advance_free
    
    rational_gcd_loop:
        xorl EXP, EXP
        div TEMP
        movl TEMP, VAL
        movl EXP, TEMP
    rational_gcd_loop_begin:
        testl TEMP, TEMP
        jnz rational_gcd_loop

        movl VAL, UNEV
        popl VAL
        cltd
        idivl UNEV
        movl VAL, TEMP
        popl VAL
        cmpl VAL, UNEV
        je rational_integer
        cltd
        idivl UNEV

make_simplified_rational:
        movl VAL, 4(FREE)
        movl (FREE), VAL
        movl VAL, 8(FREE)
        movl VAL, 16(FREE)
        orl $256, 16(FREE)    #This will be the RATIONAL header
        movl FREE, VAL        #Now VAL is the INTEGER denominator
        movl TEMP, 12(FREE)
        leal 8(FREE), UNEV    #Now UNEV is the INTEGER numerator
        addl $24, FREE
        movl FREE, -4(FREE)
        movl UNEV, (FREE)
        movl VAL, 4(FREE)
        leal -8(FREE), VAL    #Now VAL is the RATIONAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_numerator:
        call get_last_number
        cmpw $INTEGER, TEMP
        jne numerator_rational
        ret
    numerator_rational:
        cmpw $RATIONAL, TEMP
        jne error_expected_rational
        movl 4(VAL), VAL
        movl (VAL), VAL
        ret
##------------------------------------------------------------------------------
prim_denominator:
        call get_last_number
        cmpw $INTEGER, TEMP
        jne denominator_rational
        movl TEMP, (FREE)
        movl $1, 4(FREE)
        movl FREE, VAL
        jmp advance_free
    denominator_rational:
        cmpw $RATIONAL, TEMP
        jne error_expected_rational
        movl 4(VAL), VAL
        movl 4(VAL), VAL
        ret
##------------------------------------------------------------------------------
prim_quotient: 
        movl $INTEGER, (FREE)
        testl ARGL, ARGL
        jz error_too_few_args
        call get_integer_exactness
        movl 4(EXP), VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        call get_integer_exactness
        movl 4(EXP), TEMP
        cltd
        idivl TEMP
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_remainder: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl 4(EXP), VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        call get_integer_exactness
        movl 4(EXP), TEMP
        cltd
        idivl TEMP
        movl EXP, VAL
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_modulo: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl 4(EXP), VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        call get_integer_exactness
        movl 4(EXP), TEMP
        cltd
        idivl TEMP
        testl VAL, VAL
        movl EXP, VAL
        jge return_with_given_header
        addl TEMP, VAL
        jmp return_with_given_header
##------------------------------------------------------------------------------
    negate:
        negl VAL
        jo error_overflow
        jmp return_with_given_header
prim_abs:
        call insure_one_last_arg
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl 4(EXP), VAL
        test VAL, VAL
        js negate
        mov EXP, VAL
        ret
##------------------------------------------------------------------------------
    prim_gcd_zero:
        xorl VAL, VAL
        jmp return_with_given_header

prim_gcd:
        movl $INTEGER, (FREE)
        testl ARGL, ARGL
        jz prim_gcd_zero
        call get_integer_exactness
        movl 4(EXP), VAL
        testl VAL, VAL
        jge prim_gcd_next
        negl VAL
        jo error_overflow
    prim_gcd_next:
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz prim_gcd_end
        call get_integer_exactness
        movl 4(EXP), TEMP
        testl TEMP, TEMP
        jge prim_gcd_loop_begin
        negl TEMP
        jo error_overflow
        jmp prim_gcd_loop_begin

    prim_gcd_loop:
        xorl EXP, EXP
        div TEMP
        movl TEMP, VAL
        movl EXP, TEMP
    prim_gcd_loop_begin:
        test TEMP, TEMP
        jnz prim_gcd_loop
        jmp prim_gcd_next

    prim_gcd_end:
        xorl EXP, EXP         #Clear Scheme register
        jmp return_with_given_header
##------------------------------------------------------------------------------
    prim_lcm_one:
        movl $1, VAL
        jmp return_with_given_header

prim_lcm:
        movl $INTEGER, (FREE)
        testl ARGL, ARGL
        jz prim_lcm_one
        call get_integer_exactness
        movl 4(EXP), VAL
        testl VAL, VAL
        jge prim_lcm_next
        negl VAL
        jo error_overflow
    prim_lcm_next:
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz prim_lcm_end
        call get_integer_exactness
        movl 4(EXP), TEMP
        testl TEMP, TEMP
        jge prim_lcm_product
        negl TEMP
        jo error_overflow
    prim_lcm_product:
        pushl VAL
        mull TEMP
        movl VAL, UNEV
        popl VAL
        pushl UNEV
        pushl EXP
        jmp prim_lcm_loop_begin

    prim_lcm_loop:
        xorl EXP, EXP
        div TEMP
        movl TEMP, VAL
        movl EXP, TEMP
    prim_lcm_loop_begin:
        test TEMP, TEMP
        jnz prim_lcm_loop
        movl VAL, UNEV
        popl EXP
        popl VAL
        divl UNEV
        jmp prim_lcm_next

    prim_lcm_end:
        xorl EXP, EXP         #Clear Scheme register
        xorl UNEV, UNEV       #Clear Scheme register
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_min: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl EXP, VAL         #Min thus far
    min_loop: 
        testl ARGL, ARGL
        jz min_max_end
        call get_integer_exactness
        movl 4(EXP), TEMP
        cmpl 4(VAL), TEMP
        jnl min_not_less
        movl EXP, VAL
    min_not_less:
        movl 4(ARGL), ARGL     #ARGL <- (cdr ARGL)
        jmp min_loop

min_max_end:
        movl 4(VAL), VAL
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_max: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl $INTEGER, (FREE)
        call get_integer_exactness
        movl EXP, VAL         #Max thus far
    max_loop: 
        testl ARGL, ARGL
        jz min_max_end
        call get_integer_exactness
        movl 4(EXP), TEMP
        cmpl 4(VAL), TEMP
        jng max_not_greater
        movl EXP, VAL
    max_not_greater:
        movl 4(ARGL), ARGL     #ARGL <- (cdr ARGL)
        jmp max_loop
##------------------------------------------------------------------------------
prim_exact_to_inexact:
        call get_last_number
        orl $0x10000, TEMP
        jmp prim_return_new_exactness
##------------------------------------------------------------------------------
prim_inexact_to_exact:
        call get_last_number
        andl $0xffff, TEMP
    prim_return_new_exactness:
        cmpw $RATIONAL, TEMP
        je new_exactness_rational
        movl TEMP, (FREE)
        movl 4(VAL), TEMP
        movl TEMP, 4(FREE)
        movl FREE, VAL
        jmp advance_free
    new_exactness_rational:
        xorw $256, TEMP
        movl TEMP, (FREE)
        movl 4(VAL), VAL
        movl (VAL), TEMP
        movl 4(VAL), VAL
        movl 4(TEMP), TEMP
        movl 4(VAL), VAL
        jmp make_simplified_rational
##------------------------------------------------------------------------------
prim_round:
        call get_last_number
        cmpw $INTEGER, TEMP
        je rational_to_integer_end
        call rational_to_integer
        shll EXP       #Double the remainder
        testl EXP, EXP
        js prim_round_negative
        cmpl TEMP, EXP #compare the doubled remainder with the denominator
        jl return_with_given_header
        je prim_round_to_even
        incl VAL
        jmp return_with_given_header
    prim_round_to_even:
        andb $0xfe, VAL
        jmp return_with_given_header
    prim_round_negative:
        negl EXP
        jo error_overflow
        cmpl TEMP, EXP #compare the doubled remainder with the denominator
        jl return_with_given_header
        je prim_round_to_even
        decl VAL
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_ceiling:
        call get_last_number
        cmpw $INTEGER, TEMP
        je rational_to_integer_end
        call rational_to_integer
        testl EXP, EXP
        js return_with_given_header
        incl VAL
        jmp return_with_given_header
##------------------------------------------------------------------------------
rational_to_integer:  #Rational in VAL, header in TEMP
        cmpw $RATIONAL, TEMP
        jne error_expected_rational
        movw $INTEGER, TEMP
        movl TEMP, (FREE)
        movl 4(VAL), VAL      #VAL is now (numerator . denominator)
        movl 4(VAL), TEMP      #TEMP is now denominator
        movl (TEMP), EXP
        orl EXP, (FREE)
        movl 4(TEMP), TEMP
        movl (VAL), VAL       #VAL is now numerator
        movl (VAL), EXP
        orl EXP, (FREE)
        movl 4(VAL), VAL
        cltd
        idivl TEMP
    rational_to_integer_end:
        ret
##------------------------------------------------------------------------------
prim_floor:
        call get_last_number
        cmpw $INTEGER, TEMP
        je rational_to_integer_end
        call rational_to_integer
        testl EXP, EXP
        jns return_with_given_header
        decl VAL
        jmp return_with_given_header
##------------------------------------------------------------------------------
prim_truncate:
        call get_last_number
        cmpw $INTEGER, TEMP
        je rational_to_integer_end
        call rational_to_integer
        cltd
        idivl TEMP
        jmp return_with_given_header
##------------------------------------------------------------------------------
##PREDICATE PRIMITIVES
##------------------------------------------------------------------------------
prim_not: 
        call insure_one_last_arg
        cmpl $false, (ARGL)
        je return_true
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_eq_pred: 
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
    eq_loop: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz return_true
        cmpl (ARGL), EXP
        je eq_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_eqv_pred: 
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
    eqv_loop: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), VAL        #VAL <- (car ARGL)
        cmpl EXP, VAL
        je eqv_loop             #same object
        testl VAL, VAL
        jz prim_eqv_pred_false
        testl EXP, EXP
        jz prim_eqv_pred_false
        movl (VAL), TEMP       #TEMP <- (car VAL)
        cmpl (EXP), TEMP
        jne prim_eqv_pred_false
        testl $1, TEMP        #Pairs are not allowed to be eqv? unless eq?
        jz prim_eqv_pred_false
        cmpw $VECTOR, TEMP    #Vectors are not allowed to be eqv? unless eq?
        je prim_eqv_pred_false
        movl 4(VAL), TEMP       #TEMP <- (cdr VAL)
        cmpl 4(EXP), TEMP
        je eqv_loop
    prim_eqv_pred_false: 
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_null_pred: 
        call insure_one_last_arg
        cmpl $0, (ARGL)
        je return_true
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_eof_object_pred: 
        call insure_one_last_arg
        movl (ARGL), TEMP
        jecxz return_false
        cmpl $EOF, (TEMP)
        jne return_false
        movl $true, VAL
        ret
##------------------------------------------------------------------------------
prim_complex_pred:
        movl $COMPLEX, VAL
        jmp test_type_complexity
##------------------------------------------------------------------------------
prim_real_pred:
        movl $REAL, VAL
        jmp test_type_complexity
##------------------------------------------------------------------------------
prim_rational_pred:
        movl $RATIONAL, VAL
        jmp test_type_complexity
##------------------------------------------------------------------------------
prim_integer_pred:
        movl $INTEGER, VAL
test_type_complexity:
        call insure_one_last_arg
        movl (ARGL), TEMP
        jecxz return_false
        movl (TEMP), TEMP
        cmpb VAL, TEMP
        jne return_false
        cmpb VALH, TEMPH
        ja return_false
        jmp return_true
##------------------------------------------------------------------------------
prim_exact_pred:
        call insure_one_last_arg
        call get_number
        testl $0x10000, TEMP
        jz return_true
        jmp return_false
##------------------------------------------------------------------------------
prim_inexact_pred:
        call insure_one_last_arg
        call get_number
        testl $0x10000, TEMP
        jnz return_true
        jmp return_false
##------------------------------------------------------------------------------
test_major_type: 
        call insure_one_last_arg
        movl (ARGL), TEMP
        jecxz return_false
        movl (TEMP), TEMP
        cmpb VAL, TEMP
        jnz return_false
return_true: 
        movl $true, VAL
        ret
##------------------------------------------------------------------------------
return_false: 
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_pair_pred: 
        call insure_one_last_arg
        movl (ARGL), TEMP
        jecxz return_false
        testl $1, (TEMP)
        jnz return_false
        movl $true, VAL
        ret
##------------------------------------------------------------------------------
prim_symbol_pred: 
        movl $SYMBOL, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_number_pred: 
        movl $NUMBER, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_procedure_pred: 
        movl $PROCEDURE, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_boolean_pred: 
        movl $BOOLEAN, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_char_pred: 
        movl $CHAR, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_string_pred: 
        movl $STRING, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_input_port_pred: 
        movl $INPUT_PORT, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_output_port_pred: 
        movl $OUTPUT_PORT, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_vector_pred: 
        movl $VECTOR, VAL
        jmp test_major_type
##------------------------------------------------------------------------------
prim_zero_pred: 
        call get_last_number
        cmpw $INTEGER, TEMP
        jne return_false
        cmpl $0, 4(VAL)
        jz return_true
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_positive_pred: 
        call get_last_number
        cmpw $INTEGER, TEMP
        je positive_integer
    positive_rational:
        cmpw $RATIONAL, TEMP
        jne error_expected_rational
        movl 4(VAL), VAL
        movl (VAL), VAL
    positive_integer:
        cmpl $0, 4(VAL)
        jg return_true
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_negative_pred: 
        call get_last_number
        cmpw $INTEGER, TEMP
        je negative_integer
    negative_rational:
        cmpw $RATIONAL, TEMP
        jne error_expected_rational
        movl 4(VAL), VAL
        movl (VAL), VAL
    negative_integer:
        cmpl $0, 4(VAL)
        jl return_true
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_odd_pred: 
        call insure_one_last_arg
        call get_integer_exactness
        testl $1, 4(EXP)
        jz return_false
        movl $true, VAL
        ret
##------------------------------------------------------------------------------
prim_even_pred: 
        call insure_one_last_arg
        call get_integer_exactness
        testl $1, 4(EXP)
        jnz return_false
        movl $true, VAL
        ret
##------------------------------------------------------------------------------
prim_equals: 
        movl $compare_equals, thunk
compare_integers: 
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_integer
        movl (EXP), TEMP
        cmpw $INTEGER, TEMP
        jne error_expected_integer
    compare_integers_loop: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz return_true
        movl 4(EXP), VAL
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_integer
        movl (EXP), TEMP
        cmpw $INTEGER, TEMP
        jne error_expected_integer
        cmpl 4(EXP), VAL
        jmpl *thunk

compare_equals: 
        je compare_integers_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_greater: 
        movl $compare_greater, thunk
        jmp compare_integers
compare_greater: 
        jg compare_integers_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_less: 
        movl $compare_less, thunk
        jmp compare_integers
compare_less: 
        jl compare_integers_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_greater_or_equals: 
        movl $compare_greater_or_equals, thunk
        jmp compare_integers
compare_greater_or_equals: 
        jge compare_integers_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_less_or_equals: 
        movl $compare_less_or_equals, thunk
        jmp compare_integers
compare_less_or_equals: 
        jle compare_integers_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_equals: 
        movl $compare_chars_equals, thunk
compare_chars: 
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_char
        movl (EXP), TEMP
        cmpb $CHAR, TEMP
        jne error_expected_char
    compare_chars_loop: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_char
        movl (EXP), VAL
        cmpb $CHAR, VAL
        jne error_expected_char
        cmpb VALH, TEMPH
        jmpl *thunk

compare_chars_equals: 
        je compare_chars_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_greater: 
        movl $compare_chars_greater, thunk
        jmp compare_chars
compare_chars_greater: 
        ja compare_chars_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_less: 
        movl $compare_chars_less, thunk
        jmp compare_chars
compare_chars_less: 
        jb compare_chars_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_greater_or_equals: 
        movl $compare_chars_greater_or_equals, thunk
        jmp compare_chars
compare_chars_greater_or_equals: 
        jae compare_chars_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_less_or_equals: 
        movl $compare_chars_less_or_equals, thunk
        jmp compare_chars
compare_chars_less_or_equals: 
        jbe compare_chars_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_ci_equals: 
        movl $compare_chars_ci_equals, thunk
compare_chars_ci: 
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_char
        movl (EXP), TEMP
        cmpb $CHAR, TEMP
        jne error_expected_char
        cmpb $'A', TEMPH
        jb compare_chars_ci_loop
        cmpb $'Z', TEMPH
        ja compare_chars_ci_loop
        addb $32, TEMPH
    compare_chars_ci_loop: 
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz return_true
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_char
        movl (EXP), VAL
        cmpb $CHAR, VAL
        jne error_expected_char
        cmpb $'A', VALH
        jb compare_chars_ci_compare
        cmpb $'Z', VALH
        ja compare_chars_ci_compare
        addb $32, VALH
    compare_chars_ci_compare:
        cmpb VALH, TEMPH
        jmpl *thunk

compare_chars_ci_equals: 
        je compare_chars_ci_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_ci_greater: 
        movl $compare_chars_ci_greater, thunk
        jmp compare_chars_ci
compare_chars_ci_greater: 
        ja compare_chars_ci_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_ci_less: 
        movl $compare_chars_ci_less, thunk
        jmp compare_chars_ci
compare_chars_ci_less: 
        jb compare_chars_ci_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_ci_greater_or_equals: 
        movl $compare_chars_ci_greater_or_equals, thunk
        jmp compare_chars_ci
compare_chars_ci_greater_or_equals: 
        jae compare_chars_ci_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
prim_char_ci_less_or_equals: 
        movl $compare_chars_ci_less_or_equals, thunk
        jmp compare_chars_ci
compare_chars_ci_less_or_equals: 
        jbe compare_chars_ci_loop
        movl $false, VAL
        ret
##------------------------------------------------------------------------------
compare_strings_equals: 
        jne compare_strings_false #Differing length automatically disqualifies
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
    compare_strings_equals_loop: 
        cmpsb
        jne compare_strings_false
        loopw compare_strings_equals_loop
        jmp compare_strings_next

prim_string_equals: 
        movl $compare_strings_equals, thunk
compare_strings: 
        testl ARGL, ARGL
        jz return_true
        movl ARGL, EXP
        movl (EXP), UNEV
        testl UNEV, UNEV
        jz error_expected_string
        movl (UNEV), VAL
        cmpb $STRING, VAL
        jne error_expected_string
        shrl $16, VAL           #ax now holds length of first string
        movl 4(UNEV), UNEV       #UNEV now points to first character
    compare_strings_loop: 
        movl 4(EXP), EXP
        testl EXP, EXP
        jz compare_strings_true
        movl (EXP), ARGL
        testl ARGL, ARGL
        jz error_expected_string
        movl (ARGL), TEMP
        cmpb $STRING, TEMP
        jne error_expected_string
        shrl $16, TEMP           #cx now holds length of second string
        movl 4(ARGL), ARGL       #ARGL now points to first character
        movl ARGL, ENV          #remember second string for next pair
        cmpw TEMP, VAL
        jae compare_strings_longer
        movw VAL, TEMP            #Use shorter string
        ##jmp *thunk
    compare_strings_longer: 
        movw TEMP, VAL            #remember length of shorter string
        jmp *thunk

    compare_strings_next: 
        movl ENV, UNEV  #second string becomes first
        jmp compare_strings_loop

    compare_strings_true: 
        movl $true, VAL
        jmp compare_strings_end

    compare_strings_false: 
        movl $false, VAL
    compare_strings_end: 
        xorl UNEV, UNEV  #Clear scheme register
        xorl ARGL, ARGL  #Clear scheme register
        xorl ENV, ENV  #Clear scheme register
        ret
##------------------------------------------------------------------------------
prim_string_greater: 
        movl $compare_strings_greater, thunk
        jmp compare_strings

compare_strings_greater: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_false
        incw TEMP                #compare the null character as well
    compare_strings_greater_loop: 
        cmpsb
        ja compare_strings_next
        jb compare_strings_false
        loopw compare_strings_greater_loop
        jmp compare_strings_false
##------------------------------------------------------------------------------
prim_string_less: 
        movl $compare_strings_less, thunk
        jmp compare_strings

compare_strings_less: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_false
        incw TEMP                #compare the null character as well
    compare_strings_less_loop: 
        cmpsb
        jb compare_strings_next
        ja compare_strings_false
        loopw compare_strings_less_loop
        jmp compare_strings_false
##------------------------------------------------------------------------------
prim_string_less_or_equals: 
        movl $compare_strings_less_or_equals, thunk
        jmp compare_strings

compare_strings_less_or_equals: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
    compare_strings_less_or_equals_loop: 
        cmpsb
        ja compare_strings_false
        jb compare_strings_next
        loopw compare_strings_less_or_equals_loop
        jmp compare_strings_next
##------------------------------------------------------------------------------
prim_string_greater_or_equals: 
        movl $compare_strings_greater_or_equals, thunk
        jmp compare_strings

compare_strings_greater_or_equals: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
    compare_strings_greater_or_equals_loop: 
        cmpsb
        jb compare_strings_false
        ja compare_strings_next
        loopw compare_strings_greater_or_equals_loop
        jmp compare_strings_next
##------------------------------------------------------------------------------
compare_string_byte_ci: 
        lodsb
        movb (ARGL), EXP
        incl ARGL
        cmpb $'a', VAL
        jb compare_string_byte_ci_dl
        cmpb $'z', VAL
        ja compare_string_byte_ci_dl
        subb $32, VAL
    compare_string_byte_ci_dl: 
        cmpb $'a', EXP
        jb compare_string_byte_ci_cmp
        cmpb $'z', EXP
        ja compare_string_byte_ci_cmp
        subb $32, EXP
    compare_string_byte_ci_cmp: 
        cmpb EXP, VAL
        ret
##------------------------------------------------------------------------------
prim_string_ci_equals: 
        movl $compare_strings_ci_equals, thunk
        jmp compare_strings

compare_strings_ci_equals: 
        jne compare_strings_false #Differing length automatically disqualifies
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
        pushl VAL
        pushl EXP
    compare_strings_ci_equals_loop: 
        call compare_string_byte_ci
        jne compare_strings_ci_false
        loopw compare_strings_ci_equals_loop
compare_strings_ci_next: 
        popl EXP
        popl VAL
        jmp compare_strings_next

compare_strings_ci_false: 
        popl EXP
        addl $4, SP
        jmp compare_strings_false
##------------------------------------------------------------------------------
prim_string_ci_greater: 
        movl $compare_strings_ci_greater, thunk
        jmp compare_strings

compare_strings_ci_greater: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_false
        incw TEMP                #compare the null character as well
        pushl VAL
        pushl EXP
    compare_strings_ci_greater_loop: 
        call compare_string_byte_ci
        ja compare_strings_ci_next
        loopw compare_strings_ci_greater_loop
        jmp compare_strings_ci_false
##------------------------------------------------------------------------------
prim_string_ci_less: 
        movl $compare_strings_ci_less, thunk
        jmp compare_strings

compare_strings_ci_less: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_false
        incw TEMP                #compare the null character as well
        pushl VAL
        pushl EXP
    compare_strings_ci_less_loop: 
        call compare_string_byte_ci
        jb compare_strings_ci_next
        loopw compare_strings_ci_less_loop
        jmp compare_strings_ci_false
##------------------------------------------------------------------------------
prim_string_ci_less_or_equals: 
        movl $compare_strings_ci_less_or_equals, thunk
        jmp compare_strings

compare_strings_ci_less_or_equals: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
        pushl VAL
        pushl EXP
    compare_strings_ci_less_or_equals_loop: 
        call compare_string_byte_ci
        ja compare_strings_ci_false
        loopw compare_strings_ci_less_or_equals_loop
        jmp compare_strings_ci_next
##------------------------------------------------------------------------------
prim_string_ci_greater_or_equals: 
        movl $compare_strings_ci_greater_or_equals, thunk
        jmp compare_strings

compare_strings_ci_greater_or_equals: 
        cmpl ARGL, UNEV          #Same string?
        je compare_strings_next
        incw TEMP                #compare the null character as well
        pushl VAL
        pushl EXP
    compare_strings_ci_greater_or_equals_loop: 
        call compare_string_byte_ci
        jb compare_strings_ci_false
        loopw compare_strings_ci_greater_or_equals_loop
        jmp compare_strings_ci_next
##------------------------------------------------------------------------------
prim_char_alphabetic_pred:
        call get_last_char_ascii
        cmpb $'a', TEMP
        jb prim_char_upper_case_test
        cmpb $'z', TEMP
        ja return_false
        jmp return_true
##------------------------------------------------------------------------------
prim_char_lower_case_pred:
        call get_last_char_ascii
        cmpb $'a', TEMP
        jb return_false
        cmpb $'z', TEMP
        ja return_false
        jmp return_true
##------------------------------------------------------------------------------
prim_char_upper_case_pred:
        call get_last_char_ascii
    prim_char_upper_case_test:
        cmpb $'A', TEMP
        jb return_false
        cmpb $'Z', TEMP
        ja return_false
        jmp return_true
##------------------------------------------------------------------------------
prim_char_numeric_pred:
        call get_last_char_ascii
        cmpb $'0', TEMP
        jb return_false
        cmpb $'9', TEMP
        ja return_false
        jmp return_true
##------------------------------------------------------------------------------
prim_char_whitespace_pred:
        call get_last_char_ascii
        cmpb $' ', TEMP
        je return_true
        cmpb $9, TEMP    #tab
        je return_true
        cmpb $10, TEMP   #line-feed
        je return_true
        cmpb $12, TEMP   #form-feed
        je return_true
        cmpb $13, TEMP   #carriage-return
        je return_true
        jmp return_false
##------------------------------------------------------------------------------
##LIST PRIMITIVES
##------------------------------------------------------------------------------
prim_cons: 
        movl ARGL, VAL
        testl VAL, VAL
        jz error_too_few_args
        movl 4(ARGL), ARGL       #ARGL <- (cdr ARGL)
        call insure_one_last_arg
        movl (ARGL), ARGL        #ARGL <- (car ARGL)
        movl ARGL, 4(VAL)       #(set-cdr! VAL ARGL)
        ret
##------------------------------------------------------------------------------
prim_car: 
        call insure_one_last_arg
        movl (ARGL), ARGL
    car:
        call insure_pair
        movl (ARGL), VAL
        ret
##------------------------------------------------------------------------------
prim_cdr: 
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdr:
        call insure_pair
        movl 4(ARGL), VAL
        ret
##------------------------------------------------------------------------------
##Get ready!  Here they come!
##------------------------------------------------------------------------------
prim_caar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caar:
        call insure_pair
        movl (ARGL), ARGL
        jmp car
##------------------------------------------------------------------------------
prim_cadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp car
        ret
##------------------------------------------------------------------------------
prim_cdar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cdr
##------------------------------------------------------------------------------
prim_cddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cdr
##------------------------------------------------------------------------------
prim_caaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp caar
##------------------------------------------------------------------------------
prim_caadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp caar
##------------------------------------------------------------------------------
prim_cadar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cadar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cadr
##------------------------------------------------------------------------------
prim_caddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cadr
##------------------------------------------------------------------------------
prim_cdaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cdar
##------------------------------------------------------------------------------
prim_cdadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cdar
##------------------------------------------------------------------------------
prim_cddar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cddar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cddr
##------------------------------------------------------------------------------
prim_cdddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cddr
##------------------------------------------------------------------------------
prim_caaaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caaaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp caaar
##------------------------------------------------------------------------------
prim_caaadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caaadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp caaar
##------------------------------------------------------------------------------
prim_caadar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caadar:
        call insure_pair
        movl (ARGL), ARGL
        jmp caadr
##------------------------------------------------------------------------------
prim_caaddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caaddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp caadr
##------------------------------------------------------------------------------
prim_cadaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cadaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cadar
##------------------------------------------------------------------------------
prim_cadadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cadadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cadar
##------------------------------------------------------------------------------
prim_caddar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    caddar:
        call insure_pair
        movl (ARGL), ARGL
        jmp caddr
##------------------------------------------------------------------------------
prim_cadddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cadddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp caddr
##------------------------------------------------------------------------------
prim_cdaaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdaaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cdaar
##------------------------------------------------------------------------------
prim_cdaadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdaadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cdaar
##------------------------------------------------------------------------------
prim_cdadar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdadar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cdadr
##------------------------------------------------------------------------------
prim_cdaddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdaddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cdadr
##------------------------------------------------------------------------------
prim_cddaar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cddaar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cddar
##------------------------------------------------------------------------------
prim_cddadr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cddadr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cddar
##------------------------------------------------------------------------------
prim_cdddar:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cdddar:
        call insure_pair
        movl (ARGL), ARGL
        jmp cdddr
##------------------------------------------------------------------------------
prim_cddddr:
        call insure_one_last_arg
        movl (ARGL), ARGL
    cddddr:
        call insure_pair
        movl 4(ARGL), ARGL
        jmp cdddr
##------------------------------------------------------------------------------
prim_set_car:                    #ARGL should be (pair obj)
        testl ARGL, ARGL
        jz error_too_few_args
        movl 4(ARGL), VAL       #VAL should be (obj)
        movl (ARGL), ARGL        #ARGL should be pair
        call insure_pair
        testl VAL, VAL
        jz error_too_few_args
        cmpl $0, 4(VAL)
        jne error_too_many_args
        movl (VAL), VAL        #VAL is obj
        movl VAL, (ARGL)
        ret
##------------------------------------------------------------------------------
prim_set_cdr:                    #ARGL should be (pair obj)
        testl ARGL, ARGL
        jz error_too_few_args
        movl 4(ARGL), VAL       #VAL should be (obj)
        movl (ARGL), ARGL        #ARGL <- (car ARGL)
        call insure_pair
        testl VAL, VAL
        jz error_too_few_args
        cmpl $0, 4(VAL)
        jne error_too_many_args
        movl (VAL), VAL        #VAL is obj
        movl VAL, 4(ARGL)       #(set-cdr! ARGL VAL)
        ret
##------------------------------------------------------------------------------
prim_list: 
        movl ARGL, VAL
        ret
##------------------------------------------------------------------------------
prim_reverse: 
        call insure_one_last_arg
        movl (ARGL), ARGL
        testl ARGL, ARGL
        jz prim_list
        testl $1, (ARGL)
        jnz error_expected_pair
reverse: 
        xorl VAL, VAL
        jmp prim_reverse_begin
    prim_reverse_loop: 
        movl (ARGL), TEMP
        movl TEMP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        movl 4(ARGL), ARGL
        call advance_free
    prim_reverse_begin: 
        testl ARGL, ARGL
        jnz prim_reverse_loop
    prim_reverse_end: 
    prim_append_end:
        ret
##------------------------------------------------------------------------------
    prim_append_null:
        xorl VAL, VAL
        ret
    prim_append_simple:
        movl EXP, VAL
        ret
prim_append:
        testl ARGL, ARGL
        jz prim_append_null
    prim_append_recurse:
        movl (ARGL), EXP
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz prim_append_simple
        save EXP
        call prim_append_recurse
        restore EXP
append: #Appends VAL to end of list in EXP, then returns list in VAL.
        testl EXP, EXP
        jz prim_append_end
        testl $1, (EXP)
        jnz error_expected_pair
        movl (EXP), TEMP
        save TEMP
        movl 4(EXP), EXP
        call append
        restore EXP
        movl EXP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_length: 
        call insure_one_last_arg
        movl (ARGL), ARGL
        xorl TEMP, TEMP
        testl ARGL, ARGL
        jz prim_length_end
        testl $1, (ARGL)
        jnz error_expected_pair
    prim_length_loop: 
        incl TEMP
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jnz prim_length_loop

    prim_length_end: 
        movl $INTEGER, (FREE)
        movl TEMP, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_list_tail:
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL
        movl 4(ARGL), ARGL
        call get_last_exact_integer
        movl 4(TEMP), TEMP
        jmp prim_list_tail_loop_begin

    prim_list_tail_loop:
        testl VAL, VAL
        jz error_expected_pair
        testl $1, (VAL)
        jnz error_expected_pair
        movl 4(VAL), VAL
    prim_list_tail_loop_begin:
        decl TEMP
        jns prim_list_tail_loop
        ret
##------------------------------------------------------------------------------
prim_list_ref:
        call prim_list_tail
        testl VAL, VAL
        jz error_expected_pair
        testl $1, (VAL)
        jnz error_expected_pair
        movl (VAL), VAL
        ret
##------------------------------------------------------------------------------
prim_memv:
        movl $1, (thunk)
        jmp prim_mem
prim_memq:
        movl $0, (thunk)
    prim_mem:
        testl ARGL, ARGL
        jz error_too_few_args
        movl 0(ARGL), VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        movl 0(ARGL), ARGL
        jmp prim_mem_begin
    prim_mem_loop: 
        movl 4(ARGL), ARGL
    prim_mem_begin:
        testl ARGL, ARGL
        jz return_false
        testl $1, ARGL
        jnz error_expected_pair
        movl (ARGL), EXP
        cmpl EXP, VAL
        je prim_mem_found
        cmpl $0, (thunk)
        je prim_mem_loop         ##memq gives up here
        testl VAL, VAL
        jz prim_mem_loop
        testl EXP, EXP
        jz prim_mem_loop
        movl (VAL), TEMP
        cmpl (EXP), TEMP
        jne prim_mem_loop
        testl $1, TEMP        #Pairs are not allowed to be eqv? unless eq?
        jz prim_mem_loop
        cmpw $VECTOR, TEMP    #Vectors are not allowed to be eqv? unless eq?
        je prim_mem_loop
        movl 4(VAL), TEMP
        cmpl 4(EXP), TEMP
        jne prim_mem_loop
    prim_mem_found:
        movl ARGL, VAL
        ret
##------------------------------------------------------------------------------
prim_assv:
        movl $1, (thunk)
        jmp prim_ass
prim_assq:
        movl $0, (thunk)
    prim_ass:
        testl ARGL, ARGL
        jz error_too_few_args
        movl 0(ARGL), VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        movl 0(ARGL), ARGL
        jmp prim_ass_begin
    prim_ass_loop: 
        movl 4(ARGL), ARGL
    prim_ass_begin:
        testl ARGL, ARGL
        jz return_false
        testl $1, ARGL
        jnz error_expected_pair
        movl (ARGL), EXP
        testl $1, EXP
        jnz error_expected_pair
        movl (EXP), EXP
        cmpl EXP, VAL
        je prim_ass_found
        cmpl $0, (thunk)
        je prim_ass_loop         ##assq gives up here
        testl VAL, VAL
        jz prim_ass_loop
        testl EXP, EXP
        jz prim_ass_loop
        movl (VAL), TEMP
        cmpl (EXP), TEMP
        jne prim_ass_loop
        testl $1, TEMP        #Pairs are not allowed to be eqv? unless eq?
        jz prim_ass_loop
        cmpw $VECTOR, TEMP    #Vectors are not allowed to be eqv? unless eq?
        je prim_ass_loop
        movl 4(VAL), TEMP
        cmpl 4(EXP), TEMP
        jne prim_ass_loop
    prim_ass_found:
        movl (ARGL), VAL
        ret
##------------------------------------------------------------------------------
##CHAR PRIMITIVES
##------------------------------------------------------------------------------
prim_char_to_integer: 
        call get_last_char_ascii
        movl $INTEGER, (FREE)
        movl TEMP, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_integer_to_char: 
        call get_last_exact_integer
        movl 4(TEMP), TEMP
    return_char: #Returns CHAR whose ascii is in TEMP
        andl $0xff, TEMP
        leal chararray(, TEMP, 4), VAL
        ret
##------------------------------------------------------------------------------
prim_char_upcase:
        call get_last_char_ascii
        cmpb $'a', TEMP
        jb return_char
        cmpb $'z', TEMP
        ja return_char
        subb $32, TEMP
        jmp return_char
##------------------------------------------------------------------------------
prim_char_downcase:
        call get_last_char_ascii
        cmpb $'A', TEMP
        jb return_char
        cmpb $'Z', TEMP
        ja return_char
        addb $32, TEMP
        jmp return_char
##------------------------------------------------------------------------------
##STRING AND VECTOR PRIMITIVES
##------------------------------------------------------------------------------
prim_string_ref: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), EXP #String argument
        testl EXP, EXP
        jz error_expected_string
        movl 4(EXP), VAL       #VAL now points to first character of string
        movl (EXP), EXP
        cmpb $STRING, EXP
        jne error_expected_string
        movl 4(ARGL), ARGL
        call get_last_exact_integer
        movl 4(TEMP), TEMP       #index
        shrl $16, EXP           #EXP now holds length of string
        cmpl EXP, TEMP
        jae error_invalid_index
        xorl EXP, EXP      #Clear scheme register
        addl TEMP, VAL
        movb (VAL), VAL
        andl $0xff, VAL
        lea chararray(, VAL, 4), VAL
        ret
##------------------------------------------------------------------------------
prim_string_set: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), EXP #String argument
        testl EXP, EXP
        jz error_expected_string
        movl 4(EXP), VAL       #VAL now points to first character of string
        movl (EXP), EXP
        cmpw $MUTABLE_STRING, EXP
        jne error_expected_mutable_string
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz error_too_few_args
        call get_exact_integer
        movl 4(TEMP), TEMP       #index
        shrl $16, EXP           #EXP now holds length of string
        cmpl EXP, TEMP
        jae error_invalid_index
        xorl EXP, EXP      #Clear scheme register
        addl TEMP, VAL
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        movl (ARGL), TEMP #Char argument
        testl TEMP, TEMP
        jz error_expected_char
        movl (TEMP), TEMP
        cmpb $CHAR, TEMP
        jne error_expected_char
        movb TEMPH, (VAL)
        xorl VAL, VAL          #Clear scheme register
        ret
##------------------------------------------------------------------------------
prim_vector_ref: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL #Vector argument
        testl VAL, VAL
        jz error_expected_vector
        movl (VAL), EXP
        cmpw $VECTOR, EXP
        jne error_expected_vector
        movl 4(ARGL), ARGL
        call get_last_exact_integer
        movl 4(TEMP), TEMP       #index
        shrl $16, EXP           #EXP now holds length of vector
        cmpl EXP, TEMP
        jae error_invalid_index
        xorl EXP, EXP          #Clear scheme register
        movl 4(VAL, TEMP, 4), VAL
        ret
##------------------------------------------------------------------------------
prim_vector_set: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), VAL #Vector argument
        testl VAL, VAL
        jz error_expected_vector
        movl (VAL), EXP
        cmpw $VECTOR, EXP
        jne error_expected_vector
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz error_too_few_args
        call get_exact_integer
        movl 4(ARGL), ARGL
        call insure_one_last_arg
        movl 4(TEMP), TEMP       #index
        shrl $16, EXP           #EXP now holds length of vector
        cmpl EXP, TEMP
        jae error_invalid_index
        lea 4(VAL, TEMP, 4), EXP
        movl (ARGL), VAL
        movl VAL, (EXP)
        ret
##------------------------------------------------------------------------------
prim_vector_length: 
        call insure_one_last_arg
        movl (ARGL), TEMP #Vector argument
        testl TEMP, TEMP
        jz error_expected_vector
        movl (TEMP), TEMP
        cmpw $VECTOR, TEMP
        jne error_expected_vector
        jmp return_length
##------------------------------------------------------------------------------
prim_string_length: 
        call get_last_string
return_length: 
        shrl $16, TEMP         #TEMP now holds length of string or vector
        movl $INTEGER, (FREE)
        movl TEMP, 4(FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_string_to_symbol: 
        call get_last_string
        movl 4(ARGL), UNEV
        movl freesymbol, ARGL
        cld
        movl $SYMBOL, (ARGL)
        addl $4, ARGL
    prim_string_to_symbol_loop: 
        movsb
        cmpb $0, -1(UNEV)
        jne prim_string_to_symbol_loop
        jmp read_name_end       #UNEV will get cleared here
##------------------------------------------------------------------------------
prim_symbol_to_string: 
        call insure_one_last_arg
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_symbol
        movl (VAL), TEMP
        cmpb $SYMBOL, TEMP
        jne error_expected_symbol
        leal 4(VAL), TEMP
        movl TEMP, 4(FREE)
        xorl VAL, VAL
    prim_symbol_to_string_loop: 
        cmpb $0, (TEMP)
        je prim_symbol_to_string_end
        incl TEMP
        incl VAL
        jmp prim_symbol_to_string_loop

    prim_symbol_to_string_end: 
        shll $16, VAL
        movw $IMMUTABLE_STRING, VAL
        movl VAL, (FREE)
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_vector: 
        movl ARGL, VAL
        jmp list_to_vector
##------------------------------------------------------------------------------
prim_list_to_vector: 
        call insure_one_last_arg
        movl (ARGL), VAL
        testl VAL, VAL
        jz list_to_vector
        testl $1, (VAL)
        jnz error_expected_pair
list_to_vector: 
        movl VAL, UNEV  #We will step through the list using UNEV
        movl FREE, TEMP  #Start of vector
    list_to_vector_loop: 
        testl UNEV, UNEV
        jz list_to_vector_end
        addl $4, FREE
        cmpl memlimit, FREE
        jae list_to_vector_gc
        movl (UNEV), EXP
        movl EXP, (FREE)
        movl 4(UNEV), UNEV
        jmp list_to_vector_loop

    list_to_vector_gc: 
        subl TEMP, FREE
        pushl FREE              #remember how much room we had
        call gc_sans_strings
        movl memlimit, TEMP
        subl FREE, TEMP
        cmpl (SP), TEMP        #Do we have more room now?
        jbe error_out_of_memory
        addl $4, SP
        jmp list_to_vector

    list_to_vector_end: 
        movl FREE, VAL
        subl TEMP, VAL
        shll $14, VAL   #divide by four and shift to high word
        movw $VECTOR, VAL
        movl VAL, (TEMP)
        movl TEMP, VAL
        addl $4, FREE
        testl $7, FREE
        jz list_to_vector_aligned
        movl $0, (FREE)
        addl $4, FREE
    list_to_vector_aligned: 
        ret
##------------------------------------------------------------------------------
prim_make_string: 
        testl ARGL, ARGL
        jz error_too_few_args
        call get_exact_integer
        movl 4(ARGL), ARGL
        movl TEMP, EXP
        movl 4(EXP), TEMP       #TEMP is length of new string
        testl TEMP, TEMP
        js error_invalid_index
        call insure_string_space
        cld
        testl ARGL, ARGL
        jnz prim_make_string_fill
        movl freestring, ARGL
        movl 4(EXP), TEMP
        addl TEMP, ARGL
        jmp prim_make_string_finish

    prim_make_string_fill: 
        call insure_one_last_arg
        movl (ARGL), TEMP
        testl TEMP, TEMP
        jz error_expected_char
        movl (TEMP), TEMP
        cmpb $CHAR, TEMP
        jne error_expected_char
        movb TEMPH, VALL
        movl freestring, ARGL
        movl 4(EXP), TEMP
        rep
        stosb

    prim_make_string_finish: 
        xorb VAL, VAL
        stosb           #End string 
        movl freestring, VAL
        movl ARGL, freestring
        movl 4(EXP), TEMP
        shll $16, TEMP
        movw $MUTABLE_STRING, TEMP
        movl TEMP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        xorl ARGL, ARGL  #Clear scheme register
        jmp advance_free
##------------------------------------------------------------------------------
    prim_rational_to_string:
        pushl (VAL)    #Save numerator for later
        movl 4(VAL), VAL
        movl 4(VAL), VAL
        call prim_integer_to_string
        popl TEMP
        pushl VAL       #Stringified denominator
        movl 4(TEMP), VAL
        call prim_integer_to_string
        popl TEMP       #Stringified denominator
        movl TEMP, (FREE)
        movl $0, 4(FREE)
        movl FREE, ARGL
        call advance_free #ARGL <- (denominator)
        movl ARGL, 4(FREE)
        movl $string_slash, (FREE)
        movl FREE, ARGL
        call advance_free #ARGL <- ("/" denominator)
        movl VAL, (FREE)
        movl ARGL, 4(FREE)
        movl FREE, ARGL
        call advance_free #ARGL <- (numerator "/" denominator)
        jmp prim_string_append
        
prim_number_to_string: 
        testl ARGL, ARGL
        jz error_too_few_args
        call get_number
        movl $10, radix
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz prim_number_to_string_continue
        call get_last_exact_integer
        movl 4(TEMP), TEMP
        movl TEMP, radix
    prim_number_to_string_continue: 
        movl (VAL), TEMP
        movl 4(VAL), VAL
        cmpw $RATIONAL, TEMP
        je prim_rational_to_string
    prim_integer_to_string:
        call stringify_number
        pushl TEMP              #Pointer to string data
        negl TEMP
        addl $outputend, TEMP   #Now TEMP is length of string
        pushl TEMP
        call insure_string_space
        popl TEMP
        movl TEMP, ARGL
        shll $16, ARGL
        movw $MUTABLE_STRING, ARGL
        movl ARGL, (FREE)
        movl freestring, ARGL
        movl ARGL, 4(FREE)
        popl UNEV
        incl TEMP
        rep
        movsb
        movl ARGL, freestring
        xorl ARGL, ARGL  #Clear Scheme register
        xorl UNEV, UNEV  #Clear Scheme register
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_string_to_number: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), UNEV
        testl UNEV, UNEV
        jz error_expected_string
        movl (UNEV), TEMP
        cmpb $STRING, TEMP
        jne error_expected_string
        movl $10, radix          #Default radix
        movl 4(ARGL), ARGL
        testl ARGL, ARGL
        jz prim_string_to_number_continue
        call get_last_exact_integer
        movl 4(TEMP), TEMP
        movl TEMP, radix
    prim_string_to_number_continue: 
        movl $prim_string_to_number_stream, thunk
        movl 4(UNEV), UNEV                       #UNEV is string data
        pushw input                             #remember existing stream
        call prim_string_to_number_stream       #Prime the new stream
        call digitize_stream
        popw input                              #restore stream
        testl EXP, EXP
        jnz prim_string_to_number_failure
        movl $INTEGER, (FREE)
        jecxz prim_string_to_number_integer
        cmpb $'/', TEMP
        jne prim_string_to_number_failure
        movl VAL, 4(FREE)
        pushw input
        call prim_string_to_number_stream
        call digitize_stream
        popw input
        testl EXP, EXP
        jnz prim_string_to_number_failure
        jecxz prim_string_to_number_rational
    prim_string_to_number_failure: 
        xorl UNEV, UNEV  #Clear Scheme register
        xorl EXP, EXP  #Clear Scheme register
        movl $false, VAL
        ret

    prim_string_to_number_integer: 
        xorl UNEV, UNEV  #Clear Scheme register
        movl VAL, 4(FREE)
        movl FREE, VAL
        jmp advance_free

    prim_string_to_number_rational:
        movl 4(FREE), TEMP
        jmp make_rational

prim_string_to_number_stream: 
        movb (UNEV), TEMP
        movb TEMP, input
        incl UNEV
        ret
##------------------------------------------------------------------------------
prim_make_vector: 
        testl ARGL, ARGL
        jz error_too_few_args
        call get_exact_integer
        movl 4(ARGL), ARGL
        movl $false, VAL       #Default fill
        testl ARGL, ARGL
        jz prim_make_vector_continue
        cmpl $0, 4(ARGL)
        jnz error_too_many_args
        movl (ARGL), VAL        #VAL is fill
    prim_make_vector_continue: 
        movl 4(TEMP), TEMP
        pushl TEMP              #TEMP is length of new vector
        leal 4(FREE, TEMP, 4), TEMP
        cmpl memlimit, TEMP      #Will we have enough room?
        jb prim_make_vector_fill
        call gc_sans_strings
        movl (SP), TEMP
        leal 4(FREE, TEMP, 4), TEMP
        cmpl memlimit, TEMP      #We ought to have enough room now.
        jae error_out_of_memory
    prim_make_vector_fill: 
        movl (SP), TEMP
        shll $16, TEMP
        movw $VECTOR, TEMP
        movl FREE, ARGL
        movl TEMP, (ARGL)
        addl $4, ARGL
        popl TEMP
        rep
        stosl
        testl $7, ARGL
        jz prim_make_vector_aligned
        xorl VAL, VAL
        stosl
    prim_make_vector_aligned: 
        movl FREE, VAL
        movl ARGL, FREE
        xorl ARGL, ARGL  #Clear Scheme register
        ret
##------------------------------------------------------------------------------
prim_string_append: 
        xorl TEMP, TEMP
        movl ARGL, UNEV
        jmp prim_string_append_length_begin

    prim_string_append_length_loop: 
        movl (UNEV), VAL
        testl VAL, VAL
        jz error_expected_string
        movl (VAL), VAL
        cmpb $STRING, VAL
        jne error_expected_string
        shrl $16, VAL
        addl VAL, TEMP
        movl 4(UNEV), UNEV
    prim_string_append_length_begin: 
        testl UNEV, UNEV
        jnz prim_string_append_length_loop

        xorl VAL, VAL
        call insure_string_space
        shll $16, TEMP
        movw $MUTABLE_STRING, TEMP
        movl TEMP, (FREE)
        movl ARGL, VAL
        movl freestring, ARGL
        movl ARGL, 4(FREE)
        jmp prim_string_append_concat_begin

    prim_string_append_concat_loop: 
        movl (VAL), UNEV
        movl (UNEV), TEMP
        movl 4(UNEV), UNEV
        shrl $16, TEMP
        rep
        movsb
        movl 4(VAL), VAL
    prim_string_append_concat_begin: 
        testl VAL, VAL
        jnz prim_string_append_concat_loop

        movb $0, (ARGL)
        movl ARGL, freestring
        xorl UNEV, UNEV
        xorl ARGL, ARGL
        movl FREE, VAL
        jmp advance_free
##------------------------------------------------------------------------------
prim_vector_to_list:
        call insure_one_last_arg
        movl (ARGL), EXP
        testl EXP, EXP
        jz error_expected_vector
        movl (EXP), TEMP
        cmpw $VECTOR, TEMP
        jne error_expected_vector
    vector_to_list: #Vector in EXP, list returned in VAL
        xorl VAL, VAL
        shrl $16, TEMP  #Now TEMP holds vector length
        jz vector_to_list_end
    vector_to_list_loop:
        movl TEMP, backup
        movl (EXP, TEMP, 4), TEMP      #(vector-ref (- TEMP 1) EXP)
        movl TEMP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        call advance_free
        movl backup, TEMP
        decw TEMP
        jnz vector_to_list_loop
    vector_to_list_end:
        ret
##------------------------------------------------------------------------------
##CONTROL PROCEDURES
##------------------------------------------------------------------------------
prim_for_each: 
        movl $true, TEMP
        jmp prim_map_begin
prim_map: 
        movl $0, TEMP
    prim_map_begin: 
        testl ARGL, ARGL
        jz error_too_few_args
        movl (ARGL), EXP        #proc
        testl EXP, EXP
        jz error_expected_procedure
        save TEMP
        movl 4(ARGL), ARGL       #list of args
        testl ARGL, ARGL
        jz error_too_few_args
    prim_map_list_loop: 
        save EXP
        xorl VAL, VAL          #will be cars
        xorl UNEV, UNEV          #will be cdrs
    prim_map_loop: 
        movl (ARGL), EXP        #current list
        testl EXP, EXP
        jz prim_map_end
        movl (EXP), TEMP
        testl $1, TEMP
        jnz error_expected_pair
        movl TEMP, (FREE)
        movl UNEV, 4(FREE)
        movl FREE, UNEV
        call advance_free
        movl 4(EXP), TEMP
        movl TEMP, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        call advance_free
        movl 4(ARGL), ARGL       #next list
        testl ARGL, ARGL
        jnz prim_map_loop
        movl VAL, EXP          #so it will not get clobbered
        movl UNEV, ARGL
        call reverse
        movl VAL, UNEV          #current args
        movl EXP, ARGL
        call reverse            #VAL is now next list of lists
        movl root, EXP
        movl (EXP), EXP        #peek proc
        save VAL
        movl EXP, VAL
        call prim_apply_direct
        restore ARGL
        restore EXP
        restore UNEV
        cmpl $true, UNEV
        je prim_for_each_continue
        movl VAL, (FREE)
        movl UNEV, 4(FREE)
        movl FREE, UNEV          #UNEV <- (cons VAL UNEV)
        call advance_free
    prim_for_each_continue: 
        save UNEV
        jmp prim_map_list_loop

    prim_map_end: 
        movl root, TEMP
        movl 4(TEMP), TEMP
        movl TEMP, root
        restore ARGL
        cmpl $true, ARGL        #was this a for-each?
        jne reverse
        movl ARGL, VAL
        ret
##------------------------------------------------------------------------------
prim_call_with_cc: 
        call insure_one_last_arg
        xorl UNEV, UNEV
        movl SP, TEMP
    prim_cc_loop: 
        cmpl stackbase, TEMP
        je prim_cc_stack_copied
        pushl TEMP
        movl (TEMP), TEMP
        movl TEMP, 4(FREE)
        movl $SPECIAL_FORM, (FREE)
        movl FREE, VAL
        addl $8, FREE
        movl UNEV, 4(FREE)
        movl VAL, (FREE)
        movl FREE, UNEV
        call advance_free
        popl TEMP
        addl $4, TEMP
        jmp prim_cc_loop

    prim_cc_stack_copied: 
        save ENV
        movl root, VAL
        movl VAL, (FREE)
        movl 4(VAL), TEMP
        movl TEMP, root          #pop
        movl UNEV, 4(FREE)
        movl FREE, VAL          #VAL is (root . stack)
        addl $8, FREE
        movl VAL, 4(FREE)
        movl $CONTINUATION, (FREE)
        movl FREE, UNEV
        call advance_free
        movl $0, 4(FREE)
        movl UNEV, (FREE)
        movl FREE, UNEV
        call advance_free
        movl (ARGL), VAL
        testl VAL, VAL
        jz error_expected_procedure
        jmp apply_procedure
##------------------------------------------------------------------------------
##BIT OPERATIONS
##------------------------------------------------------------------------------
return_exact_integer:
        movl $INTEGER, (FREE)
        movl VAL, 4(FREE)
        movl FREE, VAL
        xorl EXP, EXP
        jmp advance_free

prim_bitwise_or: 
        xorl VAL, VAL
    bitwise_or_loop: 
        testl ARGL, ARGL
        jz return_exact_integer
        call get_exact_integer
        orl 4(TEMP), VAL
        jo error_overflow
        movl 4(ARGL), ARGL     #edi <- (cdr edi)
        jmp bitwise_or_loop
##------------------------------------------------------------------------------
prim_bitwise_xor: 
        xorl VAL, VAL
    bitwise_xor_loop: 
        testl ARGL, ARGL
        jz return_exact_integer
        call get_exact_integer
        xorl 4(TEMP), VAL
        jo error_overflow
        movl 4(ARGL), ARGL     #edi <- (cdr edi)
        jmp bitwise_xor_loop
##------------------------------------------------------------------------------
prim_bitwise_and: 
        movl $-1, VAL
    bitwise_and_loop: 
        testl ARGL, ARGL
        jz return_exact_integer
        call get_exact_integer
        andl 4(TEMP), VAL
        jo error_overflow
        movl 4(ARGL), ARGL     #edi <- (cdr edi)
        jmp bitwise_and_loop
##------------------------------------------------------------------------------
prim_bitwise_not:
        call get_last_exact_integer
        movl 4(TEMP), VAL
        notl VAL
        jmp return_exact_integer
.data
##------------------------------------------------------------------------------
msg_too_many_args:              .asciz "Too many arguments."
msg_too_few_args:               .asciz "Too few arguments."
msg_undefined_var:              .asciz " is not defined."
msg_expected_procedure:         .asciz "Expected procedure."
msg_expected_pair:              .asciz "Expected pair."
msg_expected_input_port:        .asciz "Expected input port."
msg_expected_output_port:       .asciz "Expected output port."
msg_expected_rational:          .asciz "Expected rational."
msg_expected_integer:           .asciz "Expected integer."
msg_expected_exact_integer:     .asciz "Expected exact integer."
msg_expected_number:            .asciz "Expected number."
msg_expected_char:              .asciz "Expected char."
msg_expected_string:            .asciz "Expected string."
msg_expected_mutable_string:    .asciz "Expected mutable string."
msg_expected_vector:            .asciz "Expected vector."
msg_expected_symbol:            .asciz "Expected symbol."
msg_expected_simple_type:       .asciz "Expected simple type."
msg_expected_paren:             .asciz "Expected closing parenthesis."
msg_extra_eol:                  .asciz "Too many closing parentheses."
msg_invalid_pair:               .asciz "Invalid pair."
msg_invalid_index:              .asciz "Invalid index."
msg_out_of_memory:              .asciz "Out of memory."
msg_overflow:                   .asciz "Arithmetic overflow."
msg_unsupported_rational:       .asciz "Non-integral result unsupported."
msg_io:                         .asciz "IO error."
msg_unknown_literal:            .asciz "Unknown literal."
msg_eof:                        .asciz "EOF encountered."

msg_primitive:                  .asciz "primitive "
msg_special:                    .asciz "special "
msg_closure:                    .asciz "closure "
msg_continuation:               .asciz "continuation "

msg_welcome:                    .ascii "Dream version 1.6, "
                                .asciz "Copyright (C) 2004 David Joseph Stith"
prompt:                         .byte 10
                                .asciz "dream> "
percent_s:                      .asciz "%s"
O_RDONLY:                       .asciz "r"
O_WRCRTR:                       .asciz "w"

.align 4
dream_false:
false:                          .long  BOOLEAN_FALSE
dream_true:
true:                           .long  BOOLEAN_TRUE
eof:                            .long  EOF
string_port:                    .long  INPUT_PORT
                                .long  string_port_pointer
string_port_pointer:            .long  init_scheme

sym_eol:                        .long  SYMBOL
                                .asciz  ")"
.align 4
string_slash:                   .long IMMUTABLE_STRING + 0x10000
                                .long slash
slash:                          .asciz "/"
.align 4
sym_space_name:                 .long  SYMBOL
                                .asciz  "space"
.align 4
sym_dot:                        .long  SYMBOL
                                .asciz  "."
.align 4
sym_define:                     .long  MEMOIZABLE_SYMBOL
                                .asciz  "define"
.align 4
sym_quote:                      .long  MEMOIZABLE_SYMBOL
                                .asciz  "quote"
.align 4
sym_quasiquote:                 .long  MEMOIZABLE_SYMBOL
                                .asciz  "quasiquote"
.align 4
sym_unquote:                    .long  MEMOIZABLE_SYMBOL
                                .asciz  "unquote"
.align 4
sym_unquote_splicing:           .long  MEMOIZABLE_SYMBOL
                                .asciz  "unquote-splicing"
.align 4
sym_set:                        .long  MEMOIZABLE_SYMBOL
                                .asciz  "set!"
.align 4
sym_lambda:                     .long  MEMOIZABLE_SYMBOL
                                .asciz  "lambda"
.align 4
sym_begin:                      .long  MEMOIZABLE_SYMBOL
                                .asciz  "begin"
.align 4
sym_if:                         .long  MEMOIZABLE_SYMBOL
                                .asciz  "if"
.align 4
sym_cond:                       .long  MEMOIZABLE_SYMBOL
                                .asciz  "cond"
.align 4
sym_let:                        .long  MEMOIZABLE_SYMBOL
                                .asciz  "let"
.align 4
sym_let_nested:                 .long  MEMOIZABLE_SYMBOL
                                .asciz  "let*"
.align 4
sym_letrec:                     .long  MEMOIZABLE_SYMBOL
                                .asciz  "letrec"
.align 4
sym_case:                       .long  MEMOIZABLE_SYMBOL
                                .asciz  "case"
.align 4
sym_else:                       .long  MEMOIZABLE_SYMBOL
                                .asciz  "else"
.align 4
sym_arrow:                      .long  MEMOIZABLE_SYMBOL
                                .asciz  "=>"
.align 4
sym_and:                        .long  MEMOIZABLE_SYMBOL
                                .asciz  "and"
.align 4
sym_or:                         .long  MEMOIZABLE_SYMBOL
                                .asciz  "or"
##------------------------------------------------------------------------------
##Technically the R4RS Standard requires the following to not be memoizable
##------------------------------------------------------------------------------
.ifdef MEMOIZE_BUILTINS
.equ    BUILTIN_SYMBOL, MEMOIZABLE_SYMBOL
.else
.equ    BUILTIN_SYMBOL, SYMBOL
.endif
.align 4
sym_current_input_port:         .long  BUILTIN_SYMBOL
                                .asciz  "current-input-port"
.align 4
sym_current_output_port:        .long  BUILTIN_SYMBOL
                                .asciz  "current-output-port"
.align 4
sym_apply:                      .long  BUILTIN_SYMBOL
                                .asciz  "apply"
.align 4
sym_read:                       .long  BUILTIN_SYMBOL
                                .asciz  "read"
.align 4
sym_read_char:                  .long  BUILTIN_SYMBOL
                                .asciz  "read-char"
.align 4
sym_peek_char:                  .long  BUILTIN_SYMBOL
                                .asciz  "peek-char"
.align 4
sym_write:                      .long  BUILTIN_SYMBOL
                                .asciz  "write"
.align 4
sym_newline:                    .long  BUILTIN_SYMBOL
                                .asciz  "newline"
.align 4
sym_write_char:                 .long  BUILTIN_SYMBOL
                                .asciz  "write-char"
.align 4
sym_display:                    .long  BUILTIN_SYMBOL
                                .asciz  "display"
.align 4
sym_load:                       .long  BUILTIN_SYMBOL
                                .asciz  "load"
.align 4
sym_open_input_file:            .long  BUILTIN_SYMBOL
                                .asciz  "open-input-file"
.align 4
sym_close_input_port:           .long  BUILTIN_SYMBOL
                                .asciz  "close-input-port"
.align 4
sym_open_output_file:           .long  BUILTIN_SYMBOL
                                .asciz  "open-output-file"
.align 4
sym_close_output_port:          .long  BUILTIN_SYMBOL
                                .asciz  "close-output-port"
.align 4
sym_exit:                       .long  BUILTIN_SYMBOL
                                .asciz  "exit"
.align 4
sym_eq_pred:                    .long  BUILTIN_SYMBOL
                                .asciz  "eq?"
.align 4
sym_eqv_pred:                   .long  BUILTIN_SYMBOL
                                .asciz  "eqv?"
.align 4
sym_min:                        .long  BUILTIN_SYMBOL
                                .asciz  "min"
.align 4
sym_max:                        .long  BUILTIN_SYMBOL
                                .asciz  "max"
.align 4
sym_gcd:                        .long  BUILTIN_SYMBOL
                                .asciz  "gcd"
.align 4
sym_lcm:                        .long  BUILTIN_SYMBOL
                                .asciz  "lcm"
.align 4
sym_floor:                      .long  BUILTIN_SYMBOL
                                .asciz  "floor"
.align 4
sym_ceiling:                    .long  BUILTIN_SYMBOL
                                .asciz  "ceiling"
.align 4
sym_round:                      .long  BUILTIN_SYMBOL
                                .asciz  "round"
.align 4
sym_truncate:                   .long  BUILTIN_SYMBOL
                                .asciz  "truncate"
.align 4
sym_abs:                        .long  BUILTIN_SYMBOL
                                .asciz  "abs"
.align 4
sym_add:                        .long  BUILTIN_SYMBOL
                                .asciz  "+"
.align 4
sym_sub:                        .long  BUILTIN_SYMBOL
                                .asciz  "-"
.align 4
sym_mult:                       .long  BUILTIN_SYMBOL
                                .asciz  "*"
.align 4
sym_div:                        .long  BUILTIN_SYMBOL
                                .asciz  "/"
.align 4
sym_quotient:                   .long  BUILTIN_SYMBOL
                                .asciz  "quotient"
.align 4
sym_remainder:                  .long  BUILTIN_SYMBOL
                                .asciz  "remainder"
.align 4
sym_modulo:                     .long  BUILTIN_SYMBOL
                                .asciz  "modulo"
.align 4
sym_exact_to_inexact:           .long  BUILTIN_SYMBOL
                                .asciz  "exact->inexact"
.align 4
sym_inexact_to_exact:           .long  BUILTIN_SYMBOL
                                .asciz  "inexact->exact"
.align 4
sym_pair_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "pair?"
.align 4
sym_symbol_pred:                .long  BUILTIN_SYMBOL
                                .asciz  "symbol?"
.align 4
sym_exact_pred:                 .long  BUILTIN_SYMBOL
                                .asciz  "exact?"
.align 4
sym_inexact_pred:               .long  BUILTIN_SYMBOL
                                .asciz  "inexact?"
.align 4
sym_number_pred:                .long  BUILTIN_SYMBOL
                                .asciz  "number?"
.align 4
sym_complex_pred:               .long  BUILTIN_SYMBOL
                                .asciz  "complex?"
.align 4
sym_real_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "real?"
.align 4
sym_rational_pred:              .long  BUILTIN_SYMBOL
                                .asciz  "rational?"
.align 4
sym_integer_pred:               .long  BUILTIN_SYMBOL
                                .asciz  "integer?"
.align 4
sym_procedure_pred:             .long  BUILTIN_SYMBOL
                                .asciz  "procedure?"
.align 4
sym_boolean_pred:               .long  BUILTIN_SYMBOL
                                .asciz  "boolean?"
.align 4
sym_char_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "char?"
.align 4
sym_string_pred:                .long  BUILTIN_SYMBOL
                                .asciz  "string?"
.align 4
sym_vector_pred:                .long  BUILTIN_SYMBOL
                                .asciz  "vector?"
.align 4
sym_input_port_pred:            .long  BUILTIN_SYMBOL
                                .asciz  "input-port?"
.align 4
sym_output_port_pred:           .long  BUILTIN_SYMBOL
                                .asciz  "output-port?"
.align 4
sym_zero_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "zero?"
.align 4
sym_positive_pred:              .long  BUILTIN_SYMBOL
                                .asciz  "positive?"
.align 4
sym_negative_pred:              .long  BUILTIN_SYMBOL
                                .asciz  "negative?"
.align 4
sym_even_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "even?"
.align 4
sym_odd_pred:                   .long  BUILTIN_SYMBOL
                                .asciz  "odd?"
.align 4
sym_equals:                     .long  BUILTIN_SYMBOL
                                .asciz  "="
.align 4
sym_greater:                    .long  BUILTIN_SYMBOL
                                .asciz  ">"
.align 4
sym_less:                       .long  BUILTIN_SYMBOL
                                .asciz  "<"
.align 4
sym_greater_or_equals:          .long  BUILTIN_SYMBOL
                                .asciz  ">="
.align 4
sym_less_or_equals:             .long  BUILTIN_SYMBOL
                                .asciz  "<="
.align 4
sym_char_alphabetic_pred:       .long  BUILTIN_SYMBOL
                                .asciz  "char-alphabetic?"
.align 4
sym_char_upper_case_pred:       .long  BUILTIN_SYMBOL
                                .asciz  "char-upper-case?"
.align 4
sym_char_lower_case_pred:       .long  BUILTIN_SYMBOL
                                .asciz  "char-lower-case?"
.align 4
sym_char_whitespace_pred:       .long  BUILTIN_SYMBOL
                                .asciz  "char-whitespace?"
.align 4
sym_char_numeric_pred:          .long  BUILTIN_SYMBOL
                                .asciz  "char-numeric?"
.align 4
sym_char_equals:                .long  BUILTIN_SYMBOL
                                .asciz  "char=?"
.align 4
sym_char_ci_equals:             .long  BUILTIN_SYMBOL
                                .asciz  "char-ci=?"
.align 4
sym_char_greater:               .long  BUILTIN_SYMBOL
                                .asciz  "char>?"
.align 4
sym_char_ci_greater:            .long  BUILTIN_SYMBOL
                                .asciz  "char-ci>?"
.align 4
sym_char_less:                  .long  BUILTIN_SYMBOL
                                .asciz  "char<?"
.align 4
sym_char_ci_less:               .long  BUILTIN_SYMBOL
                                .asciz  "char-ci<?"
.align 4
sym_char_greater_or_equals:     .long  BUILTIN_SYMBOL
                                .asciz  "char>=?"
.align 4
sym_char_ci_greater_or_equals:  .long  BUILTIN_SYMBOL
                                .asciz  "char-ci>=?"
.align 4
sym_char_less_or_equals:        .long  BUILTIN_SYMBOL
                                .asciz  "char<=?"
.align 4
sym_char_ci_less_or_equals:     .long  BUILTIN_SYMBOL
                                .asciz  "char-ci<=?"
.align 4
sym_cons:                       .long  BUILTIN_SYMBOL
                                .asciz  "cons"
.align 4
sym_car:                        .long  BUILTIN_SYMBOL
                                .asciz  "car"
.align 4
sym_cdr:                        .long  BUILTIN_SYMBOL
                                .asciz  "cdr"
.align 4
sym_caar:                       .long  BUILTIN_SYMBOL
                                .asciz  "caar"
.align 4
sym_cadr:                       .long  BUILTIN_SYMBOL
                                .asciz  "cadr"
.align 4
sym_cdar:                       .long  BUILTIN_SYMBOL
                                .asciz  "cdar"
.align 4
sym_cddr:                       .long  BUILTIN_SYMBOL
                                .asciz  "cddr"
.align 4
sym_caaar:                      .long  BUILTIN_SYMBOL
                                .asciz  "caaar"
.align 4
sym_caadr:                      .long  BUILTIN_SYMBOL
                                .asciz  "caadr"
.align 4
sym_cadar:                      .long  BUILTIN_SYMBOL
                                .asciz  "cadar"
.align 4
sym_caddr:                      .long  BUILTIN_SYMBOL
                                .asciz  "caddr"
.align 4
sym_cdaar:                      .long  BUILTIN_SYMBOL
                                .asciz  "cdaar"
.align 4
sym_cdadr:                      .long  BUILTIN_SYMBOL
                                .asciz  "cdadr"
.align 4
sym_cddar:                      .long  BUILTIN_SYMBOL
                                .asciz  "cddar"
.align 4
sym_cdddr:                      .long  BUILTIN_SYMBOL
                                .asciz  "cdddr"
.align 4
sym_caaaar:                     .long  BUILTIN_SYMBOL
                                .asciz  "caaaar"
.align 4
sym_caaadr:                     .long  BUILTIN_SYMBOL
                                .asciz  "caaadr"
.align 4
sym_caadar:                     .long  BUILTIN_SYMBOL
                                .asciz  "caadar"
.align 4
sym_caaddr:                     .long  BUILTIN_SYMBOL
                                .asciz  "caaddr"
.align 4
sym_cadaar:                     .long  BUILTIN_SYMBOL
                                .asciz  "cadaar"
.align 4
sym_cadadr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cadadr"
.align 4
sym_caddar:                     .long  BUILTIN_SYMBOL
                                .asciz  "caddar"
.align 4
sym_cadddr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cadddr"
.align 4
sym_cdaaar:                     .long  BUILTIN_SYMBOL
                                .asciz  "cdaaar"
.align 4
sym_cdaadr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cdaadr"
.align 4
sym_cdadar:                     .long  BUILTIN_SYMBOL
                                .asciz  "cdadar"
.align 4
sym_cdaddr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cdaddr"
.align 4
sym_cddaar:                     .long  BUILTIN_SYMBOL
                                .asciz  "cddaar"
.align 4
sym_cddadr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cddadr"
.align 4
sym_cdddar:                     .long  BUILTIN_SYMBOL
                                .asciz  "cdddar"
.align 4
sym_cddddr:                     .long  BUILTIN_SYMBOL
                                .asciz  "cddddr"
.align 4
sym_set_car:                    .long  BUILTIN_SYMBOL
                                .asciz  "set-car!"
.align 4
sym_set_cdr:                    .long  BUILTIN_SYMBOL
                                .asciz  "set-cdr!"
.align 4
sym_list:                       .long  BUILTIN_SYMBOL
                                .asciz  "list"
.align 4
sym_reverse:                    .long  BUILTIN_SYMBOL
                                .asciz  "reverse"
.align 4
sym_append:                     .long  BUILTIN_SYMBOL
                                .asciz  "append"
.align 4
sym_length:                     .long  BUILTIN_SYMBOL
                                .asciz  "length"
.align 4
sym_list_tail:                  .long  BUILTIN_SYMBOL
                                .asciz  "list-tail"
.align 4
sym_list_ref:                   .long  BUILTIN_SYMBOL
                                .asciz  "list-ref"
.align 4
sym_memv:                       .long  BUILTIN_SYMBOL
                                .asciz  "memv"
.align 4
sym_memq:                       .long  BUILTIN_SYMBOL
                                .asciz  "memq"
.align 4
sym_assv:                       .long  BUILTIN_SYMBOL
                                .asciz  "assv"
.align 4
sym_assq:                       .long  BUILTIN_SYMBOL
                                .asciz  "assq"
.align 4
sym_eof_object_pred:            .long  BUILTIN_SYMBOL
                                .asciz  "eof-object?"
.align 4
sym_null_pred:                  .long  BUILTIN_SYMBOL
                                .asciz  "null?"
.align 4
sym_not:                        .long  BUILTIN_SYMBOL
                                .asciz  "not"
.align 4
sym_char_to_integer:            .long  BUILTIN_SYMBOL
                                .asciz  "char->integer"
.align 4
sym_integer_to_char:            .long  BUILTIN_SYMBOL
                                .asciz  "integer->char"
.align 4
sym_char_upcase:                .long  BUILTIN_SYMBOL
                                .asciz  "char-upcase"
.align 4
sym_char_downcase:              .long  BUILTIN_SYMBOL
                                .asciz  "char-downcase"
.align 4
sym_map:                        .long  BUILTIN_SYMBOL
                                .asciz  "map"
.align 4
sym_for_each:                   .long  BUILTIN_SYMBOL
                                .asciz  "for-each"
.align 4
sym_call_with_cc:               .long  BUILTIN_SYMBOL
                                .asciz  "call-with-current-continuation"
.align 4
sym_string_ref:                 .long  BUILTIN_SYMBOL
                                .asciz  "string-ref"
.align 4
sym_string_set:                 .long  BUILTIN_SYMBOL
                                .asciz  "string-set!"
.align 4
sym_string_length:              .long  BUILTIN_SYMBOL
                                .asciz  "string-length"
.align 4
sym_string_equals:              .long  BUILTIN_SYMBOL
                                .asciz  "string=?"
.align 4
sym_string_greater:             .long  BUILTIN_SYMBOL
                                .asciz  "string>?"
.align 4
sym_string_greater_or_equals:   .long  BUILTIN_SYMBOL
                                .asciz  "string>=?"
.align 4
sym_string_less:                .long  BUILTIN_SYMBOL
                                .asciz  "string<?"
.align 4
sym_string_less_or_equals:      .long  BUILTIN_SYMBOL
                                .asciz  "string<=?"
.align 4
sym_string_ci_equals:           .long  BUILTIN_SYMBOL
                                .asciz  "string-ci=?"
.align 4
sym_string_ci_greater:          .long  BUILTIN_SYMBOL
                                .asciz  "string-ci>?"
.align 4
sym_string_ci_greater_or_equals: .long BUILTIN_SYMBOL
                                .asciz  "string-ci>=?"
.align 4
sym_string_ci_less:             .long  BUILTIN_SYMBOL
                                .asciz  "string-ci<?"
.align 4
sym_string_ci_less_or_equals:   .long  BUILTIN_SYMBOL
                                .asciz  "string-ci<=?"
.align 4
sym_vector_length:              .long  BUILTIN_SYMBOL
                                .asciz  "vector-length"
.align 4
sym_string_to_symbol:           .long  BUILTIN_SYMBOL
                                .asciz  "string->symbol"
.align 4
sym_symbol_to_string:           .long  BUILTIN_SYMBOL
                                .asciz  "symbol->string"
.align 4
sym_string_to_number:           .long  BUILTIN_SYMBOL
                                .asciz  "string->number"
.align 4
sym_number_to_string:           .long  BUILTIN_SYMBOL
                                .asciz  "number->string"
.align 4
sym_make_string:                .long  BUILTIN_SYMBOL
                                .asciz  "make-string"
.align 4
sym_string_append:              .long  BUILTIN_SYMBOL
                                .asciz  "string-append"
.align 4
sym_make_vector:                .long  BUILTIN_SYMBOL
                                .asciz  "make-vector"
.align 4
sym_vector_ref:                 .long  BUILTIN_SYMBOL
                                .asciz  "vector-ref"
.align 4
sym_vector_set:                 .long  BUILTIN_SYMBOL
                                .asciz  "vector-set!"
.align 4
sym_list_to_vector:             .long  BUILTIN_SYMBOL
                                .asciz  "list->vector"
.align 4
sym_vector_to_list:             .long  BUILTIN_SYMBOL
                                .asciz  "vector->list"
.align 4
sym_vector:                     .long  BUILTIN_SYMBOL
                                .asciz  "vector"
.align 4
sym_numerator:                  .long  BUILTIN_SYMBOL
                                .asciz  "numerator"
.align 4
sym_denominator:                .long  BUILTIN_SYMBOL
                                .asciz  "denominator"
.align 4
sym_bitwise_xor:                .long   BUILTIN_SYMBOL
                                .asciz  "^"
.align 4
sym_bitwise_or:                 .long   BUILTIN_SYMBOL
                                .asciz  "|"
.align 4
sym_bitwise_and:                .long   BUILTIN_SYMBOL
                                .asciz  "&"
.align 4
sym_bitwise_not:                .long   BUILTIN_SYMBOL
                                .asciz  "~"
.align 4
sym_system:                     .long   BUILTIN_SYMBOL
                                .asciz  "system"
.ifdef RANDOM
.align 4
sym_random:                     .long   BUILTIN_SYMBOL
                                .asciz  "random"
.endif
.ifdef XLIB
.align 4
sym_x_open_window:              .long   BUILTIN_SYMBOL
                                .asciz  "x-open-window"
.align 4
sym_x_close_window:             .long   BUILTIN_SYMBOL
                                .asciz  "x-close-window"
.align 4
sym_x_clear_window:             .long   BUILTIN_SYMBOL
                                .asciz  "x-clear-window"
.align 4
sym_x_set_foreground:           .long   BUILTIN_SYMBOL
                                .asciz  "x-set-foreground"
.align 4
sym_x_flush:                    .long   BUILTIN_SYMBOL
                                .asciz  "x-flush"
.align 4
sym_x_draw_point:               .long   BUILTIN_SYMBOL
                                .asciz  "x-draw-point"
.align 4
sym_x_draw_line:                .long   BUILTIN_SYMBOL
                                .asciz  "x-draw-line"
.align 4
sym_x_draw_rectangle:           .long   BUILTIN_SYMBOL
                                .asciz  "x-draw-rectangle"
.align 4
sym_x_fill_rectangle:           .long   BUILTIN_SYMBOL
                                .asciz  "x-fill-rectangle"
.endif
.align 4
freesymbol:                     .long   symbols
radix:                          .long 10
root:                           .long 0
mem:                            .long   mem1
dream_memlimit:
memlimit:                       .long   memlimit1
memnew:                         .long   mem2
memlimitnew:                    .long   memlimit2
memstr:                         .long   memstr1
memstrlimit:                    .long   memstrlimit1
memstrnew:                      .long   memstr2
memstrlimitnew:                 .long   memstrlimit2
ext_eof:                        .asciz  "eof>"

globals:  #Try to order these from least often to most often used.
.ifdef XLIB
.long sym_x_open_window, FOREIGN_PROCEDURE, prim_x_open_window
.long sym_x_close_window, FOREIGN_PROCEDURE, prim_x_close_window
.long sym_x_clear_window, FOREIGN_PROCEDURE, prim_x_clear_window
.long sym_x_set_foreground, FOREIGN_PROCEDURE, prim_x_set_foreground
.long sym_x_flush, FOREIGN_PROCEDURE, prim_x_flush
.long sym_x_draw_point, FOREIGN_PROCEDURE, prim_x_draw_point
.long sym_x_draw_line, FOREIGN_PROCEDURE, prim_x_draw_line
.long sym_x_draw_rectangle, FOREIGN_PROCEDURE, prim_x_draw_rectangle
.long sym_x_fill_rectangle, FOREIGN_PROCEDURE, prim_x_fill_rectangle
.endif
.long sym_exit, PRIMITIVE, prim_exit
.long sym_system, PRIMITIVE, prim_system
.long sym_load, PRIMITIVE, prim_load
.long sym_caaaar, PRIMITIVE, prim_caaaar
.long sym_caaadr, PRIMITIVE, prim_caaadr
.long sym_caadar, PRIMITIVE, prim_caadar
.long sym_caaddr, PRIMITIVE, prim_caaddr
.long sym_cadaar, PRIMITIVE, prim_cadaar
.long sym_cadadr, PRIMITIVE, prim_cadadr
.long sym_caddar, PRIMITIVE, prim_caddar
.long sym_cadddr, PRIMITIVE, prim_cadddr
.long sym_cdaaar, PRIMITIVE, prim_cdaaar
.long sym_cdaadr, PRIMITIVE, prim_cdaadr
.long sym_cdadar, PRIMITIVE, prim_cdadar
.long sym_cdaddr, PRIMITIVE, prim_cdaddr
.long sym_cddaar, PRIMITIVE, prim_cddaar
.long sym_cddadr, PRIMITIVE, prim_cddadr
.long sym_cdddar, PRIMITIVE, prim_cdddar
.long sym_cddddr, PRIMITIVE, prim_cddddr
.long sym_caaar, PRIMITIVE, prim_caaar
.long sym_caadr, PRIMITIVE, prim_caadr
.long sym_cadar, PRIMITIVE, prim_cadar
.long sym_caddr, PRIMITIVE, prim_caddr
.long sym_cdaar, PRIMITIVE, prim_cdaar
.long sym_cdadr, PRIMITIVE, prim_cdadr
.long sym_cddar, PRIMITIVE, prim_cddar
.long sym_cdddr, PRIMITIVE, prim_cdddr
.long sym_caar, PRIMITIVE, prim_caar
.long sym_cadr, PRIMITIVE, prim_cadr
.long sym_cdar, PRIMITIVE, prim_cdar
.long sym_cddr, PRIMITIVE, prim_cddr
.long sym_open_input_file, PRIMITIVE, prim_open_input_file
.long sym_close_input_port, PRIMITIVE, prim_close_input_port
.long sym_open_output_file, PRIMITIVE, prim_open_output_file
.long sym_close_output_port, PRIMITIVE, prim_close_output_port
.long sym_bitwise_xor, PRIMITIVE, prim_bitwise_xor
.long sym_bitwise_or, PRIMITIVE, prim_bitwise_or
.long sym_bitwise_and, PRIMITIVE, prim_bitwise_and
.long sym_bitwise_not, PRIMITIVE, prim_bitwise_not
.long sym_numerator, PRIMITIVE, prim_numerator
.long sym_denominator, PRIMITIVE, prim_denominator
.long sym_inexact_to_exact, PRIMITIVE, prim_inexact_to_exact
.long sym_exact_to_inexact, PRIMITIVE, prim_exact_to_inexact
.long sym_exact_pred, PRIMITIVE, prim_exact_pred
.long sym_inexact_pred, PRIMITIVE, prim_inexact_pred
.long sym_string_to_symbol, PRIMITIVE, prim_string_to_symbol
.long sym_symbol_to_string, PRIMITIVE, prim_symbol_to_string
.long sym_string_to_number, PRIMITIVE, prim_string_to_number
.long sym_number_to_string, PRIMITIVE, prim_number_to_string
.long sym_string_append, PRIMITIVE, prim_string_append
.long sym_current_input_port, PRIMITIVE, prim_current_input_port
.long sym_current_output_port, PRIMITIVE, prim_current_output_port
.long sym_vector_to_list, PRIMITIVE, prim_vector_to_list
.long sym_list_to_vector, PRIMITIVE, prim_list_to_vector
.long sym_list_tail, PRIMITIVE, prim_list_tail
.long sym_list_ref, PRIMITIVE, prim_list_ref
.long sym_assv, PRIMITIVE, prim_assv
.long sym_assq, PRIMITIVE, prim_assq
.long sym_memv, PRIMITIVE, prim_memv
.long sym_memq, PRIMITIVE, prim_memq
.long sym_vector_length, PRIMITIVE, prim_vector_length
.long sym_vector_set, PRIMITIVE, prim_vector_set
.long sym_vector_ref, PRIMITIVE, prim_vector_ref
.long sym_vector, PRIMITIVE, prim_vector
.long sym_make_vector, PRIMITIVE, prim_make_vector
.long sym_append, PRIMITIVE, prim_append
.long sym_map, PRIMITIVE, prim_map
.long sym_for_each, PRIMITIVE, prim_for_each
.long sym_string_ci_greater, PRIMITIVE, prim_string_ci_greater
.long sym_string_ci_greater_or_equals, PRIMITIVE, prim_string_ci_greater_or_equals
.long sym_string_ci_less, PRIMITIVE, prim_string_ci_less
.long sym_string_ci_less_or_equals, PRIMITIVE, prim_string_ci_less_or_equals
.long sym_string_ci_equals, PRIMITIVE, prim_string_ci_equals
.long sym_make_string, PRIMITIVE, prim_make_string
.long sym_string_set, PRIMITIVE, prim_string_set
.long sym_string_ref, PRIMITIVE, prim_string_ref
.long sym_string_length, PRIMITIVE, prim_string_length
.long sym_string_greater, PRIMITIVE, prim_string_greater
.long sym_string_greater_or_equals, PRIMITIVE, prim_string_greater_or_equals
.long sym_string_less, PRIMITIVE, prim_string_less
.long sym_string_less_or_equals, PRIMITIVE, prim_string_less_or_equals
.long sym_string_equals, PRIMITIVE, prim_string_equals
.long sym_call_with_cc, PRIMITIVE, prim_call_with_cc
.long sym_eof_object_pred, PRIMITIVE, prim_eof_object_pred
.long sym_char_alphabetic_pred, PRIMITIVE, prim_char_alphabetic_pred
.long sym_char_upper_case_pred, PRIMITIVE, prim_char_upper_case_pred
.long sym_char_lower_case_pred, PRIMITIVE, prim_char_lower_case_pred
.long sym_char_whitespace_pred, PRIMITIVE, prim_char_whitespace_pred
.long sym_char_numeric_pred, PRIMITIVE, prim_char_numeric_pred
.long sym_char_equals, PRIMITIVE, prim_char_equals
.long sym_char_ci_equals, PRIMITIVE, prim_char_ci_equals
.long sym_char_greater, PRIMITIVE, prim_char_greater
.long sym_char_ci_greater, PRIMITIVE, prim_char_ci_greater
.long sym_char_less, PRIMITIVE, prim_char_less
.long sym_char_ci_less, PRIMITIVE, prim_char_ci_less
.long sym_char_greater_or_equals, PRIMITIVE, prim_char_greater_or_equals
.long sym_char_ci_greater_or_equals, PRIMITIVE, prim_char_ci_greater_or_equals
.long sym_char_less_or_equals, PRIMITIVE, prim_char_less_or_equals
.long sym_char_ci_less_or_equals, PRIMITIVE, prim_char_ci_less_or_equals
.long sym_char_upcase, PRIMITIVE, prim_char_upcase
.long sym_char_downcase, PRIMITIVE, prim_char_downcase
.long sym_char_to_integer, PRIMITIVE, prim_char_to_integer
.long sym_integer_to_char, PRIMITIVE, prim_integer_to_char
.long sym_pair_pred, PRIMITIVE, prim_pair_pred
.long sym_symbol_pred, PRIMITIVE, prim_symbol_pred
.long sym_complex_pred, PRIMITIVE, prim_complex_pred
.long sym_real_pred, PRIMITIVE, prim_real_pred
.long sym_rational_pred, PRIMITIVE, prim_rational_pred
.long sym_integer_pred, PRIMITIVE, prim_integer_pred
.long sym_number_pred, PRIMITIVE, prim_number_pred
.long sym_procedure_pred, PRIMITIVE, prim_procedure_pred
.long sym_boolean_pred, PRIMITIVE, prim_boolean_pred
.long sym_char_pred, PRIMITIVE, prim_char_pred
.long sym_vector_pred, PRIMITIVE, prim_vector_pred
.long sym_string_pred, PRIMITIVE, prim_string_pred
.long sym_input_port_pred, PRIMITIVE, prim_input_port_pred
.long sym_output_port_pred, PRIMITIVE, prim_output_port_pred
.long sym_null_pred, PRIMITIVE, prim_null_pred
.long sym_not, PRIMITIVE, prim_not
.ifdef RANDOM
.long sym_random, PRIMITIVE, prim_random
.endif
.long sym_zero_pred, PRIMITIVE, prim_zero_pred
.long sym_positive_pred, PRIMITIVE, prim_positive_pred
.long sym_negative_pred, PRIMITIVE, prim_negative_pred
.long sym_even_pred, PRIMITIVE, prim_even_pred
.long sym_odd_pred, PRIMITIVE, prim_odd_pred
.long sym_equals, PRIMITIVE, prim_equals
.long sym_greater, PRIMITIVE, prim_greater
.long sym_less, PRIMITIVE, prim_less
.long sym_greater_or_equals, PRIMITIVE, prim_greater_or_equals
.long sym_less_or_equals, PRIMITIVE, prim_less_or_equals
.long sym_set_car, PRIMITIVE, prim_set_car
.long sym_set_cdr, PRIMITIVE, prim_set_cdr
.long sym_cons, PRIMITIVE, prim_cons
.long sym_car, PRIMITIVE, prim_car
.long sym_cdr, PRIMITIVE, prim_cdr
.long sym_reverse, PRIMITIVE, prim_reverse
.long sym_length, PRIMITIVE, prim_length
.long sym_list, PRIMITIVE, prim_list
.long sym_read, PRIMITIVE, prim_read
.long sym_read_char, PRIMITIVE, prim_read_char
.long sym_peek_char, PRIMITIVE, prim_peek_char
.long sym_write, PRIMITIVE, prim_write
.long sym_display, PRIMITIVE, prim_display
.long sym_write_char, PRIMITIVE, prim_write_char
.long sym_newline, PRIMITIVE, prim_newline
.long sym_eq_pred, PRIMITIVE, prim_eq_pred
.long sym_eqv_pred, PRIMITIVE, prim_eqv_pred
.long sym_min, PRIMITIVE, prim_min
.long sym_max, PRIMITIVE, prim_max
.long sym_lcm, PRIMITIVE, prim_lcm
.long sym_gcd, PRIMITIVE, prim_gcd
.long sym_floor, PRIMITIVE, prim_floor
.long sym_ceiling, PRIMITIVE, prim_ceiling
.long sym_round, PRIMITIVE, prim_round
.long sym_truncate, PRIMITIVE, prim_truncate
.long sym_modulo, PRIMITIVE, prim_modulo
.long sym_quotient, PRIMITIVE, prim_quotient
.long sym_remainder, PRIMITIVE, prim_remainder
.long sym_abs, PRIMITIVE, prim_abs
.long sym_div, PRIMITIVE, prim_div
.long sym_mult, PRIMITIVE, prim_mult
.long sym_sub, PRIMITIVE, prim_sub
.long sym_add, PRIMITIVE, prim_add
.long sym_apply, PRIMITIVE, prim_apply
.long sym_quasiquote, SPECIAL_FORM, ev_quasiquote
.long sym_case, SPECIAL_FORM, ev_case
.long sym_or, SPECIAL_FORM, ev_or
.long sym_and, SPECIAL_FORM, ev_and
.long sym_define, SPECIAL_FORM, ev_define
.long sym_set, SPECIAL_FORM, ev_set
.long sym_letrec, SPECIAL_FORM, ev_letrec
.long sym_let_nested, SPECIAL_FORM, ev_let_nested
.long sym_let, SPECIAL_FORM, ev_let
.long sym_cond, SPECIAL_FORM, ev_cond
.long sym_quote, SPECIAL_FORM, ev_quote
.long sym_lambda, SPECIAL_FORM, ev_lambda
.long sym_begin, SPECIAL_FORM, ev_begin
.long sym_if, SPECIAL_FORM, ev_if
.long 0
init_scheme:
.ascii "(define vector=?"
.ascii   "(lambda (x y)"
.ascii     "(if (= (vector-length x) (vector-length y))"
.ascii       "(letrec"
.ascii         "((%elements=?"
.ascii           "(lambda (x y i)"
.ascii             "(if (< i 0)"
.ascii              " #t "
.ascii               "(if (equal? (vector-ref x i) (vector-ref y i))"
.ascii                 "(%elements=? x y (- i 1)) #f)))))"
.ascii         "(%elements=? x y (- (vector-length x) 1)))"
.ascii      " #f)))"
.ascii "(define equal?"
.ascii   "(lambda (x y)"
.ascii     "(if (pair? x)"
.ascii       "(and (pair? y) (equal? (car x) (car y)) (equal? (cdr x) (cdr y)))"
.ascii       "(if (string? x)"
.ascii         "(and (string? y) (string=? x y))"
.ascii         "(if (vector? x)"
.ascii           "(and (vector? y) (vector=? x y))"
.ascii           "(eqv? x y))))))"
.ascii "(define list?"
.ascii   "(lambda (lst)"
.ascii     "(define %list?"
.ascii       "(lambda (lst visited)"
.ascii         "(cond"
.ascii           "((pair? lst)"
.ascii            "(if (memq (car lst) visited)"
.ascii             " #f "
.ascii              "(%list? (cdr lst) (cons (car lst) visited))))"
.ascii           "((null? lst) #t)"
.ascii           "(else #f))))"
.ascii     "(%list? lst ())))"
.ascii "(define member"
.ascii   "(lambda (obj lst)"
.ascii     "(if (null? lst)"
.ascii      " #f "
.ascii       "(if (equal? obj (car lst))"
.ascii        " lst "
.ascii         "(member obj (cdr lst))))))"
.ascii "(define assoc"
.ascii   "(lambda (obj lst)"
.ascii     "(if (null? lst)"
.ascii      " #f "
.ascii       "(if (equal? obj (caar lst))"
.ascii         "(car lst)"
.ascii         "(assoc obj (cdr lst))))))"
.ascii "(define string->list #f)"
.ascii "(letrec"
.ascii   "((%str->list"
.ascii     "(lambda (obj lst n)"
.ascii       "(if (zero? n)"
.ascii        " lst "
.ascii         "(%str->list obj "
.ascii           "(cons (string-ref obj (- n 1)) lst)"
.ascii           "(- n 1))))))"
.ascii   "(set! string->list"
.ascii     "(lambda (str)"
.ascii       "(%str->list str () (string-length str)))))"
.ascii "(define list->string"
.ascii   "(lambda (lst)"
.ascii     "(define str (make-string (length lst)))"
.ascii     "(define fill"
.ascii       "(lambda (lst i)"
.ascii         "(if (null? lst)"
.ascii          " str "
.ascii           "(begin"
.ascii             "(string-set! str i (car lst))"
.ascii             "(fill (cdr lst) (+ i 1))))))"
.ascii     "(fill lst 0)))"
.ascii "(define string (lambda lst (list->string lst)))"
.ascii "(define substring"
.ascii   "(lambda (str start end)"
.ascii     "(define substr (make-string (- end start)))"
.ascii     "(define fill "
.ascii       "(lambda (start i)"
.ascii         "(if (= start end)"
.ascii          " substr "
.ascii           "(begin "
.ascii             "(string-set! substr i (string-ref str start))"
.ascii             "(fill (+ start 1) (+ i 1))))))"
.ascii     "(fill start 0)))"
.ascii "(define call-with-input-file"
.ascii   "(lambda (str proc)"
.ascii     "(define in (open-input-file str))"
.ascii     "(define result (proc in))"
.ascii     "(close-input-port in)"
.ascii    " result))"
.ascii "(define call-with-output-file"
.ascii   "(lambda (str proc)"
.ascii     "(define out (open-output-file str))"
.ascii     "(define result (proc out))"
.ascii     "(close-output-port out)"
.ascii    " result))"
.ifdef RATIONALS
.ascii "(let ((prim<= <=)"
.ascii       "(prim>= >=)"
.ascii       "(prim< <)"
.ascii       "(prim> >)"
.ascii       "(prim/ /)"
.ascii       "(prim* *)"
.ascii       "(prim- -)"
.ascii       "(prim+ +)"
.ascii       "(prim= =))"
.ascii   "(define loop="
.ascii     "(lambda (a x)"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(let ((b (car x)))"
.ascii           "(if (and (prim= (numerator a)"
.ascii                           "(numerator b))"
.ascii                    "(prim= (denominator a)"
.ascii                           "(denominator b)))"
.ascii               "(loop= b (cdr x))"
.ascii            " #f)))))"
.ascii   "(define loop-compare"
.ascii     "(lambda (thunk a x)"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(let ((b (car x)))"
.ascii           "(let ((cd (lcm (denominator a)"
.ascii                          "(denominator b))))"
.ascii             "(if (thunk (prim* (numerator a)"
.ascii                               "(prim/ cd (denominator a)))"
.ascii                        "(prim* (numerator b)"
.ascii                               "(prim/ cd (denominator b))))"
.ascii               "(loop-compare thunk b (cdr x))"
.ascii              " #f))))))"
.ascii   "(define loop+"
.ascii     "(lambda (a x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(let ((d (lcm (denominator a)"
.ascii                         "(denominator b))))"
.ascii             "(let ((n (prim+ (prim* (numerator a)"
.ascii                                    "(prim/ d (denominator a)))"
.ascii                             "(prim* (numerator b)"
.ascii                                    "(prim/ d (denominator b))))))"
.ascii               "(loop+"
.ascii                " (prim/ n d)"
.ascii                 "(cdr x))))))))"
.ascii   "(define loop-"
.ascii     "(lambda (a x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(let ((d (lcm (denominator a)"
.ascii                         "(denominator b))))"
.ascii             "(let ((n (prim- (prim* (numerator a)"
.ascii                                    "(prim/ d (denominator a)))"
.ascii                             "(prim* (numerator b)"
.ascii                                    "(prim/ d (denominator b))))))"
.ascii               "(loop-"
.ascii                " (prim/ n d)"
.ascii                 "(cdr x))))))))"
.ascii   "(define loop*"
.ascii     "(lambda (a x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(let ((ga (gcd (numerator a)"
.ascii                          "(denominator b)))"
.ascii                 "(gb (gcd (denominator a)"
.ascii                          "(numerator b))))"
.ascii             "(let ((n (prim* (prim/ (numerator a) ga)"
.ascii                             "(prim/ (numerator b) gb)))"
.ascii                   "(d (prim* (prim/ (denominator a) gb)"
.ascii                             "(prim/ (denominator b) ga))))"
.ascii               "(loop*"
.ascii                " (prim/ n d)"
.ascii                 "(cdr x))))))))"
.ascii   "(define loop/"
.ascii     "(lambda (a x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(let ((ga (gcd (numerator a)"
.ascii                          "(numerator b)))"
.ascii                 "(gb (gcd (denominator a)"
.ascii                          "(denominator b))))"
.ascii             "(let ((n (prim* (prim/ (numerator a) ga)"
.ascii                             "(prim/ (denominator b) gb)))"
.ascii                   "(d (prim* (prim/ (denominator a) gb)"
.ascii                             "(prim/ (numerator b) ga))))"
.ascii               "(loop/"
.ascii                " (prim/ n d)"
.ascii                 "(cdr x))))))))"
.ascii   "(set! ="
.ascii     "(lambda x"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(loop= (car x) (cdr x)))))"
.ascii   "(set! >"
.ascii     "(lambda x"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(loop-compare prim> (car x) (cdr x)))))"
.ascii   "(set! <"
.ascii     "(lambda x"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(loop-compare prim< (car x) (cdr x)))))"
.ascii   "(set! >="
.ascii     "(lambda x"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(loop-compare prim>= (car x) (cdr x)))))"
.ascii   "(set! <="
.ascii     "(lambda x"
.ascii       "(if (null? x)"
.ascii        " #t"
.ascii         "(loop-compare prim<= (car x) (cdr x)))))"
.ascii   "(set! +"
.ascii     "(lambda x"
.ascii       "(loop+ 0 x)))"
.ascii   "(set! -"
.ascii     "(lambda (f . x)"
.ascii       "(if (null? x)"
.ascii         "(/ (prim- (numerator f))"
.ascii            "(denominator f))"
.ascii         "(loop- f x))))"
.ascii   "(set! abs"
.ascii     "(lambda (x)"
.ascii       "(if (negative? x)"
.ascii         "(- x)"
.ascii        " x)))"
.ascii   "(set! min"
.ascii     "(lambda (a . x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(if (inexact? b)"
.ascii               "(set! a (exact->inexact a)))"
.ascii           "(let ((cd (lcm (denominator a)"
.ascii                          "(denominator b))))"
.ascii             "(if (prim< (prim* (numerator a)"
.ascii                               "(prim/ cd (denominator a)))"
.ascii                        "(prim* (numerator b)"
.ascii                               "(prim/ cd (denominator b))))"
.ascii               "(apply min (cons a (cdr x)))"
.ascii               "(apply min (cons b (cdr x)))))))))"
.ascii   "(set! max"
.ascii     "(lambda (a . x)"
.ascii       "(if (null? x)"
.ascii        " a"
.ascii         "(let ((b (car x)))"
.ascii           "(if (inexact? b)"
.ascii               "(set! a (exact->inexact a)))"
.ascii           "(let ((cd (lcm (denominator a)"
.ascii                          "(denominator b))))"
.ascii             "(if (prim> (prim* (numerator a)"
.ascii                               "(prim/ cd (denominator a)))"
.ascii                        "(prim* (numerator b)"
.ascii                               "(prim/ cd (denominator b))))"
.ascii               "(apply max (cons a (cdr x)))"
.ascii               "(apply max (cons b (cdr x)))))))))"
.ascii   "(set! *"
.ascii     "(lambda x"
.ascii       "(loop* 1 x)))"
.ascii   "(set! /"
.ascii     "(lambda (a . x)"
.ascii       "(if (null? x)"
.ascii         "(loop/ 1 (list a))"
.ascii         "(loop/ a x)))))"
.endif
.byte 0
##------------------------------------------------------------------------------
.bss
##------------------------------------------------------------------------------
.align 8
.lcomm mem1, MEMSIZE*1024
.lcomm memlimit1, 32*4

.align 8
.lcomm mem2, MEMSIZE*1024
.lcomm memlimit2, 32*4        

.lcomm memstr1, STRSIZE*1024
.lcomm memstrlimit1, 4

.lcomm memstr2, STRSIZE*1024
.lcomm memstrlimit2, 4

.lcomm global_environment, 4
.lcomm current_input_port, 4
.lcomm current_output_port, 4
.lcomm backup, 4
.lcomm obhash, (2<<HASHPOWER)*128
obhash_end:
.lcomm freestring, 4
.lcomm chararray, 256*4
.lcomm symbols, SYMSIZE*1024

.lcomm output, 255
.lcomm outputend, 1

.lcomm stackbase, 4
.lcomm thunk, 4
.lcomm error_continuation, 4
.lcomm io_file, 4
.lcomm dream_free, 4
.lcomm dream_error, 4
.lcomm input, 1
.lcomm backup_char, 1
